#!/usr/bin/env python3

import sys
import argparse

from AaronTools.geometry import Geometry
from AaronTools.fileIO import FileReader, read_types
from AaronTools.utils.utils import get_filename, glob_files

fukui_parser = argparse.ArgumentParser(
    description="integrate weighted Fukui functions around atoms",
    formatter_class=argparse.RawTextHelpFormatter
)

fukui_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="an FCHK file or ORCA output with MO's"
)

fukui_parser.add_argument(
    "-o", "--output",
    type=str,
    default=False,
    dest="outfile",
    help="output destination\n" +
    "$INFILE will be replaced with the name of the input file\n" +
    "Default: stdout"
)

info = fukui_parser.add_mutually_exclusive_group(required=True)
info.add_argument(
    "-fd", "--fukui-donor",
    dest="fukui_donor",
    default=False,
    action="store_true",
    help="print Fukui donor values\n"
    "see DOI 10.1002/jcc.24699 for weighting method\n"
    "for details on how the function is condensed, see the GitHub wiki:\n"
    "https://github.com/QChASM/AaronTools.py/wiki/Structure-Analysis-and-Descriptor-Implementation#condensed-fukui"
)

info.add_argument(
    "-fa", "--fukui-acceptor",
    dest="fukui_acceptor",
    default=False,
    action="store_true",
    help="print Fukui acceptor values\n"
    "see DOI 10.1021/acs.jpca.9b07516 for weighting method\n"
    "for details on how the function is condensed, see the GitHub wiki:\n"
    "https://github.com/QChASM/AaronTools.py/wiki/Structure-Analysis-and-Descriptor-Implementation#condensed-fukui"
)

info.add_argument(
    "-f2", "--fukui-dual",
    dest="fukui_dual",
    default=False,
    action="store_true",
    help="print Fukui dual values\n"
    "see DOI 10.1021/acs.jpca.9b07516 for weighting method\n"
    "for details on how the function is condensed, see the GitHub wiki:\n"
    "https://github.com/QChASM/AaronTools.py/wiki/Structure-Analysis-and-Descriptor-Implementation#condensed-fukui"
)

fukui_parser.add_argument(
    "-d", "--delta",
    type=float,
    dest="delta",
    default=0.1,
    help="delta parameter for weighting orbitals in Fukui functions\n"
    "Default: 0.1 Hartree",
)

fukui_parser.add_argument(
    "-nt", "--number-of-threads",
    type=int,
    default=1,
    dest="n_jobs",
    help="number of threads to use when evaluating basis functions\n"
    "this is on top of NumPy's multithreading,\n"
    "so if NumPy uses 8 threads and n_jobs=2, you can\n"
    "expect to see 16 threads in use\n"
    "Default: 1"
)

fukui_parser.add_argument(
    "-m", "--max-array",
    type=int,
    default=10000000,
    dest="max_length",
    help="max. array size to read from FCHK files\n"
    "a reasonable size for setting parsing orbital data\n"
    "can improve performance when reading large FCHK files\n"
    "too small of a value will prevent orbital data from\n"
    "being parsed\n"
    "Default: 10000000",
)

fukui_parser.add_argument(
    "-v", "--vdw-radii",
    default="umn",
    choices=["umn", "bondi"],
    dest="radii",
    help="VDW radii to use in calculation\n" + 
    "umn: main group vdw radii from J. Phys. Chem. A 2009, 113, 19, 5806–5812\n" +
    "    (DOI: 10.1021/jp8111556)\n" + 
    "    transition metals are crystal radii from Batsanov, S.S. Van der Waals\n" +
    "    Radii of Elements. Inorganic Materials 37, 871–885 (2001).\n" +
    "    (DOI: 10.1023/A:1011625728803)\n" + 
    "bondi: radii from J. Phys. Chem. 1964, 68, 3, 441–451 (DOI: 10.1021/j100785a001)\n" +
    "Default: umn",
)

grid_options = fukui_parser.add_argument_group("Lebedev integration options")
grid_options.add_argument(
    "-rp", "--radial-points",
    type=int,
    default=32,
    choices=[20, 32, 64, 75, 99, 127],
    dest="rpoints",
    help="number of radial shells for Gauss-Legendre integration\n" +
    "of the radial component\n" +
    "lower values are faster, but at the cost of accuracy\nDefault: 32"
)

grid_options.add_argument(
    "-ap", "--angular-points",
    type=int,
    default=302,
    choices=[110, 194, 302, 590, 974, 1454, 2030, 2702, 5810],
    dest="apoints",
    help="number of angular points for Lebedev integration\n" +
    "lower values are faster, but at the cost of accuracy\nDefault: 1454"
)

args = fukui_parser.parse_args()


infiles = glob_files(args.infile, parser=fukui_parser)
for n, f in enumerate(infiles):
    if isinstance(f, str):
        infile = FileReader(
            f, just_geom=False, max_length=args.max_length
        )
    elif len(sys.argv) >= 1:
        infile = FileReader(
            ("from stdin", "fchk", f),
            just_geom=False,
            max_length=args.max_length,
        )

    try:
        orbits = infile.other["orbitals"]
    except KeyError:
        raise RuntimeError("orbital info was not parsed from %s" % f)
    geom = Geometry(infile, refresh_connected=False, refresh_ranks=False)
    
    if args.fukui_donor:
        func = orbits.condensed_fukui_donor_values
    elif args.fukui_acceptor:
        func = orbits.condensed_fukui_acceptor_values
    elif args.fukui_dual:
        func = orbits.condensed_fukui_dual_values

    vals = func(
        geom,
        n_jobs=args.n_jobs,
        rpoints=args.rpoints,
        apoints=args.apoints,
        radii=args.radii,
    )

    s = ""
    if len(infiles) > 1:
        s += "%s:\n" % f
    for i, atom in enumerate(geom.atoms):
        s += "%s%3s\t%6.3f\n" % ("\t" if len(infiles) > 1 else "", atom.name, vals[i])

    if not args.outfile:
        print(s)
    else:
        outfile = args.outfile
        mode = "w"
        if "$INFILE" in outfile:
            outfile = outfile.replace("$INFILE", get_filename(f))
        elif n > 0:
            mode = "a"
        with open(outfile, mode) as f:
            f.write(s)

    



"""
methods that construct headers and footers can specify some keyword arguments
keywords are ORCA_*, PSI4_*, or GAUSSIAN_* (from AaronTools.theory)
ORCA_ROUTE: list(str)
ORCA_BLOCKS: dict(list(str)) - keys are block names minus %
ORCA_COORDINATES: ignored
ORCA_COMMENT: list(str)

PSI4_SETTINGS: dict(setting_name: [value])
PSI4_BEFORE_GEOM: list(str)
PSI4_AFTER_JOB: list(str) -FUNCTIONAL will be replaced with method name
PSI4_COMMENT: list(str)
PSI4_MOLECULE: dict(str:list(str)) e.g. {'symmetry': ['c1']}
PSI4_COORDINATES: dict() with keys:
                  'coords' - array of coordinates with one item for each atom
                  'variables' - list of name (str), value (float), is_angstrom (bool) tuples
                  this is ignored if using a SAPTMethod with a low-spin combination
                  of monomers
PSI4_JOB: dict(optimize/frequencies/etc: list(str -FUNCTIONAL replaced w/ method))
PSI4_OPTKING: dict(setting_name: [value])

GAUSSIAN_PRE_ROUTE: dict(list(str)) - keys are link0 minus %
GAUSSIAN_ROUTE: dict(list(str)) - e.g. {'opt': ['NoEigenTest', 'Tight']}
GAUSSIAN_COORDINATES: list of coordinates and variables/constants
GAUSSIAN_CONSTRAINTS: list(str)
GAUSSIAN_GEN_BASIS: list(str) - only filled by BasisSet automatically when writing footer
GAUSSIAN_GEN_ECP: list(str) - only filled by BasisSet automatically when writing footer
GAUSSIAN_POST: list(str)
GAUSSIAN_COMMENT: list(str)

SQM_COMMENT: list(str)
SQM_QMMM: dict()
"""

ORCA_ROUTE = "simple" # simple input
ORCA_BLOCKS = "blocks" #blocks
ORCA_COORDINATES = 3 #molecule (not used)
ORCA_COMMENT = "comments" #comments

PSI4_SETTINGS = "settings" # set { stuff }
PSI4_BEFORE_GEOM = "before_molecule" #before geometry - basis info goes here
PSI4_AFTER_JOB = "after_job" #after job
PSI4_COMMENT = "comments" #comments
PSI4_MOLECULE = "molecule" #molecule - used for symmetry etc.
PSI4_COORDINATES = "coordinates" # coordinate variables
PSI4_JOB = "job" #energy, optimize, etc
PSI4_OPTKING = "optking" #constraints go here
PSI4_BEFORE_JOB = "before_job" #before job stuff - e.g. auto_fragments
PSI4_SOLVENT = "pcm_solver" # psi4 interface to pcmsolver

GAUSSIAN_PRE_ROUTE = "link0" #can be used for things like %chk=some.chk
GAUSSIAN_ROUTE = "route" #route specifies most options, e.g. #n B3LYP/3-21G opt
GAUSSIAN_COORDINATES = "coordinates" #coordinate section
GAUSSIAN_CONSTRAINTS = "constraints" #constraints section (e.g. B 1 2 F)
GAUSSIAN_GEN_BASIS = "gen_basis" #gen or genecp basis section
GAUSSIAN_GEN_ECP = "gen_ecp" #genecp ECP section
GAUSSIAN_POST = "end_of_file" #after everything else (e.g. NBO options)
GAUSSIAN_COMMENT = "comments" #comment line after the route

SQM_COMMENT = "comments"
SQM_QMMM = "qmmm"

QCHEM_MOLECULE = "molecule" # $molecule
QCHEM_REM = "rem" # $rem
QCHEM_COMMENT = "comments" # $comment
QCHEM_SETTINGS = "section" # $*

from AaronTools.theory.basis import BasisSet, Basis, ECP
from AaronTools.theory.emp_dispersion import EmpiricalDispersion
from AaronTools.theory.implicit_solvent import ImplicitSolvent
from AaronTools.theory.grid import IntegrationGrid
from AaronTools.theory.method import Method, SAPTMethod
from AaronTools.theory.job_types import *
from AaronTools.theory.theory import Theory



#!/usr/bin/env python3

import sys
import os
import argparse

from AaronTools.ring import Ring

libaddring_parser = argparse.ArgumentParser(
    description='add a ring fragment to your personal library', \
    formatter_class=argparse.RawTextHelpFormatter
)

libaddring_parser.add_argument(
    'infile', metavar='input file',
    type=str,
    default=None,
    help='a coordinate file'
)

libaddring_parser.add_argument(
    '-n', '--name',
    type=str,
    required=False,
    default=None,
    dest='name',
    help="Name of ring being added to the library\n" +
    "if no name is given, the ring will be printed to STDOUT"
)

libaddring_parser.add_argument(
    '-w', '--walk',
    type=str,
    nargs=1,
    required=True,
    dest='walk',
    help="comma-separated list of atoms to define" +
    "the direction the ring is traversed (1-indexed)"
)

args = libaddring_parser.parse_args()

ring = Ring(args.infile)

walk_atoms = args.walk.split(',')
ring.find_end(len(walk_atoms), walk_atoms)

ring.comment = "E:%s" % args.walk

if args.name is None:
    print(ring.write(outfile=False))
else:
    ring_lib = Ring.AARON_LIBS
    ring_file = os.path.join(os.path.dirname(Ring.AARON_LIBS), args.name + '.xyz')
    if os.path.exists(ring_file):
        overwrite = input(
            "%s already exists.\nWould you like to overwrite it? (YES/no)\n" % ring_file
        )
        if overwrite != "YES":
            print("%s to overwrite, not overwriting" % overwrite)
            sys.exit(0)

    ring.write(outfile=ring_file)



#!/usr/bin/env python3

import argparse
from warnings import warn

from AaronTools.config import Config
from AaronTools.job_control import SubmitProcess
from AaronTools.utils.utils import glob_files

config = Config(quiet=True)

default_proc = config.getint(
    "Job", "processors", fallback=config.getint("Job", "procs", fallback=4)
)
default_mem = config.getint("Job", "memory", fallback=8)
default_walltime = config.getint("Job", "walltime", fallback=12)
default_template = config.get("Job", "template", fallback=None)


submit_parser = argparse.ArgumentParser(
    description="submit a QM computation to the queue",
    formatter_class=argparse.RawTextHelpFormatter,
)

submit_parser.add_argument(
    "infile",
    metavar="input file",
    type=str,
    nargs="+",
    help="a Psi4, ORCA, or Gaussian input file",
)

submit_parser.add_argument(
    "-d",
    "--config-default",
    type=str,
    default=None,
    dest="section",
    help="use memory, processors, walltime, and template from\n"
    "the specified seciton of the AaronTools config",
)

submit_parser.add_argument(
    "-j",
    "--job-template",
    type=str,
    default=None,
    dest="template",
    help="template job template file (i.e. for `qsub`, `bsub`, or `sbatch`)",
)

submit_parser.add_argument(
    "-p",
    "--processors",
    type=int,
    required=False,
    default=None,
    dest="processors",
    help="number of processors\n" "Default: %i" % default_proc,
)

submit_parser.add_argument(
    "-m",
    "--memory",
    type=int,
    required=False,
    default=None,
    dest="memory",
    help="memory in GB\n" "Default: %i" % default_mem,
)

submit_parser.add_argument(
    "-t",
    "--walltime",
    type=int,
    required=False,
    default=None,
    dest="time",
    help="walltime in hours\n" "Default: %i" % default_walltime,
)

submit_parser.add_argument(
    "-wl",
    "--wait-last",
    action="store_true",
    default=False,
    dest="wait_last",
    help="wait for the last job to finish before exiting",
)

submit_parser.add_argument(
    "-we",
    "--wait-each",
    action="store_true",
    default=False,
    dest="wait_each",
    help="wait for each job to finish before submitting the next",
)

args = submit_parser.parse_args()

for i, f in enumerate(glob_files(args.infile, parser=submit_parser)):
    # TODO: if processors etc. is not specified, read the input file to see if
    #       processors were specified

    processors = args.processors
    memory = args.memory
    walltime = args.time
    template = args.template

    if args.section is not None:
        if args.processors is None:
            processors = config.getint(
                args.section, "processors", fallback=None
            )
        if args.memory is None:
            memory = config.getint(args.section, "memory", fallback=None)
        if args.time is None:
            walltime = config.getint(args.section, "walltime", fallback=None)
        if args.template is None:
            template = config.get(args.section, "template", fallback=None)

    if processors is None:
        processors = default_proc
    if memory is None:
        memory = default_mem
    if walltime is None:
        walltime = default_walltime
    if template is None:
        template = default_template

    submit_process = SubmitProcess(
        f, walltime, processors, memory, template=template
    )

    try:
        if args.wait_each or (i == len(args.infile) - 1 and args.wait_last):
            submit_process.submit(wait=True, quiet=False)
        else:
            submit_process.submit(wait=False, quiet=False)

    except Exception as e:
        warn("failed to submit %s: %s" % (f, str(e)))



#!/usr/bin/env python3

import sys
import argparse
import os
from glob import glob
from warnings import warn

from numpy import isclose

from AaronTools.comp_output import CompOutput
from AaronTools.fileIO import FileReader
from AaronTools.geometry import Geometry
from AaronTools.utils.utils import glob_files


thermo_parser = argparse.ArgumentParser(
    description="print thermal corrections and free energy",
    formatter_class=argparse.RawTextHelpFormatter
)

thermo_parser.add_argument(
    "infile",
    metavar="frequency output file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="completed QM output file with frequency info"
)

thermo_parser.add_argument(
    "-o",
    "--output",
    type=str,
    default=None,
    required=False,
    dest="outfile",
    help="output destination \nDefault: stdout"
)

thermo_parser.add_argument(
    "-if",
    "--input-format",
    type=str,
    nargs=1,
    default=None,
    dest="input_format",
    choices=["log", "out", "dat"],
    help="file format of input - required if input is stdin"
)

thermo_parser.add_argument(
    "-sp",
    "--single-point",
    type=str,
    nargs="*",
    default=[None],
    required=False,
    dest="sp_file",
    help="file containing single-point energy"
)

thermo_parser.add_argument(
    "-t",
    "--temperature",
    type=float,
    default=None,
    required=False,
    dest="temp",
    help="compute thermal corrections using the specified temperature (K)\n" +
    "Default: value found in file or 298.15"
)

thermo_parser.add_argument(
    "-w0",
    "--frequency-cutoff",
    type=float,
    default=100.0,
    required=False,
    dest="w0",
    help="cutoff frequency for quasi free energy corrections (1/cm)\n" +
    "Default: 100 cm^-1"
)

thermo_parser.add_argument(
    "-csv",
    "--csv-format",
    nargs="?",
    required=False,
    dest="csv",
    default=False,
    choices=["comma", "semicolon", "tab", "space"],
    help="print output in CSV format with the specified delimiter"
)

thermo_parser.add_argument(
    "-r",
    "--recursive",
    metavar="PATTERN",
    type=str,
    nargs="*",
    default=None,
    required=False,
    dest="pattern",
    help="search subdirectories of current directory for files matching PATTERN"
)


args = thermo_parser.parse_args()

if args.csv is None:
    args.csv = "comma"

if args.csv:
    if args.csv == "comma":
        delim = ","
    elif args.csv == "semicolon":
        delim = ";"
    elif args.csv == "tab":
        delim = "\t"
    elif args.csv == "space":
        delim = " "

    anharm_header = delim.join([
        "E", "E+ZPE", "E+ZPE(anh)", "H(RRHO)", "G(RRHO)", "G(Quasi-RRHO)", "G(Quasi-harmonic)",
        "ZPE", "ZPE(anh)", "dH(RRHO)", "dG(RRHO)", "dG(Quasi-RRHO)", "dG(Quasi-harmonic)",
        "SP_File", "Thermo_File"
    ])
    harm_header = delim.join([
        "E", "E+ZPE", "H(RRHO)", "G(RRHO)", "G(Quasi-RRHO)", "G(Quasi-harmonic)",
        "ZPE", "dH(RRHO)", "dG(RRHO)", "dG(Quasi-RRHO)", "dG(Quasi-harmonic)",
        "SP_File", "Thermo_File"
    ])

header = None
output = ""

if args.pattern is None:
    infiles = glob_files(args.infile, parser=thermo_parser)
else:
    infiles = []
    if args.infile == [sys.stdin]:
        directories = [os.getcwd()]
    else:
        directories = []
        for directory in args.infile:
            directories.extend(glob(directory))

    for directory in directories:
        for root, dirs, files in os.walk(directory, topdown=True):
            for pattern in args.pattern:
                full_glob = os.path.join(root, pattern)
                infiles.extend(glob(full_glob))

    infiles.sort()

if args.sp_file != [None]:
    if args.pattern is None:
        sp_filenames = glob_files([f for  f in args.sp_file])

    else:
        sp_filenames = []
        if args.infile == [sys.stdin]:
            directories = [os.getcwd()]
        else:
            directories = []
            for directory in args.infile:
                directories.extend(glob(directory))

        for directory in directories:
            for root, dirs, files in os.walk(directory, topdown=True):
                for pattern in args.sp_file:
                    full_glob = os.path.join(root, pattern)
                    sp_filenames.extend(glob(full_glob))

    sp_filenames.sort()

    sp_files = [FileReader(f, just_geom=False) for f in sp_filenames]
    sp_energies = [sp_file.other["energy"] for sp_file in sp_files]

else:
    sp_energies = [None for f in infiles]
    sp_filenames = [None for f in infiles]

while len(sp_energies) < len(infiles):
    sp_energies.extend([sp_file.other["energy"] for sp_file in sp_files])
    sp_filenames.extend(args.sp_file)

while len(infiles) < len(sp_filenames):
    infiles.extend(args.infile)

for sp_nrg, sp_file, f in zip(sp_energies, sp_filenames, infiles):
    if isinstance(f, str):
        if args.input_format is not None:
            infile = FileReader((f, args.input_format[0], None), just_geom=False)
        else:
            infile = FileReader(f, just_geom=False)
    else:
        if args.input_format is not None:
            infile = FileReader(("from stdin", args.input_format[0], f), just_geom=False)
        else:
            if len(sys.argv) >= 1:
                thermo_parser.print_help()
                raise RuntimeError(
                    "when no input file is given, stdin is read and a format must be specified"
                )

    if "frequency" not in infile.other:
        warn("no frequencies in %s - skipping" % f)
        continue
    freq = infile.other["frequency"]

    co = CompOutput(
        infile,
    )

    if sp_nrg is None:
        nrg = co.energy
    else:
        nrg = sp_nrg
        sp_geom = Geometry(sp_file)
        freq_geom = Geometry(infile)
        rmsd = sp_geom.RMSD(freq_geom)
        if not isclose(rmsd, 0, atol=1e-5):
            warn(
                "\ngeometries in supposed single-point/thermochemistry pair appear\n" +
                "to be different (rmsd = %.5f)\n" % rmsd +
                "%s\n%s" % (sp_geom.name, freq_geom.name)
            )

    dE, dH, s = co.therm_corr(temperature=args.temp)
    if freq.anharm_data:
        ZPVE_anh = co.calc_zpe(anharmonic=True)
    rrho_dG = co.calc_G_corr(v0=0, temperature=args.temp, method="RRHO")
    qrrho_dG = co.calc_G_corr(v0=args.w0, temperature=args.temp, method="QRRHO")
    qharm_dG = co.calc_G_corr(v0=args.w0, temperature=args.temp, method="QHARM")

    if args.temp is None:
        t = co.temperature
    else:
        t = args.temp

    if args.csv:
        nrg_str = "%.6f" % nrg
        corrections = [co.ZPVE]
        if freq.anharm_data:
            if header != anharm_header:
                output += "%s\n" % anharm_header
                header = anharm_header
            corrections.append(ZPVE_anh)
        elif header != harm_header:
            output += "%s\n" % harm_header
            header = harm_header
        corrections.extend([dH, rrho_dG, qrrho_dG, qharm_dG])
        therm = [nrg + correction for correction in corrections]
        output += delim.join(
            [nrg_str] +
            ["%.6f" % x for x in therm] + \
            ["%.6f" % x for x in corrections] + \
            [sp_file if sp_file is not None else f, f]
        )
        output += "\n"
    else:
        output += "electronic energy of %s = %.6f Eh\n" % (
            sp_file if sp_file is not None else f, 
            nrg
        )
        output += "    E+ZPE             = %.6f Eh  (ZPE = %.6f)\n" % (nrg + co.ZPVE, co.ZPVE)
        if freq.anharm_data:
            output += "    E+ZPE(anh)        = %.6f Eh  (ZPE(anh) = %.6f)\n" % (
                nrg + ZPVE_anh, ZPVE_anh
            )
        output += "thermochemistry from %s at %.2f K:\n" % (f, t)
        output += "    H(RRHO)           = %.6f Eh  (dH = %.6f)\n" % (nrg + dH, dH)
        output += "    G(RRHO)           = %.6f Eh  (dG = %.6f)\n" % (nrg + rrho_dG, rrho_dG)
        output += "  quasi treatments for entropy (w0=%.1f cm^-1):\n" % args.w0
        output += "    G(Quasi-RRHO)     = %.6f Eh  (dG = %.6f)\n" % (nrg + qrrho_dG, qrrho_dG)
        output += "    G(Quasi-harmonic) = %.6f Eh  (dG = %.6f)\n" % (nrg + qharm_dG, qharm_dG)

        output += "\n"

output = output.strip()

if not args.outfile:
    print(output.strip())
else:
    with open(
            args.outfile,
            "a"
    ) as f:
        f.write(output.strip())



#!/usr/bin/env python3

import sys
import argparse
from warnings import warn
import numpy as np

from AaronTools.geometry import Geometry
from AaronTools.fileIO import FileReader
from AaronTools.pathway import Pathway


def width(n):
    """use to determine 0-padding based on number of files we're printing"""
    return np.ceil(np.log10(n))

instruction_args = ["-n", "-e", "-min", "-max", "-t"]

interpolate_parser = argparse.ArgumentParser(
    description="interpolate between input structures",
    formatter_class=argparse.RawTextHelpFormatter
)

interpolate_parser.add_argument(
    "infiles", metavar="infile",
    type=str,
    nargs="*",
    default=[],
    help="one or more input coordinate files"
)

interpolate_parser.add_argument(
    "-n", "--print-nstruc",
    type=int,
    nargs=1,
    default=0,
    required=False,
    dest="n_struc",
    metavar="number of structures",
    help="number of interpolated structures to print"
)

interpolate_parser.add_argument(
    "-max", "--print-maxima",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="print_max",
    help="print coordinates for potential energy maxima"
)

interpolate_parser.add_argument(
    "-min", "--print-minima",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="print_min",
    help="print coordinates for potential energy minima"
)

interpolate_parser.add_argument(
    "-e", "--print-energy",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="print_E",
    help="print energy and energy derivative instead of structures"
)

interpolate_parser.add_argument(
    "-o", "--output-destination",
    type=str,
    nargs=1,
    default=[None],
    required=False,
    metavar="output destination",
    dest="outfile",
    help="output destination\n" +
    "$i will be replaced with zero-padded numbers\n" +
    "Default: traj-$i.xyz for structures, stdout for energies"
)

interpolate_parser.add_argument(
    "-u", "--use-unfinished",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="use_incomplete",
    help="use unfinished geometries (e.g. optimization still running)"
)

interpolate_parser.add_argument(
    "-t", "--print-ts",
    type=float,
    nargs="+",
    default=False,
    required=False,
    metavar=("t1", "t2"),
    dest="specific_ts",
    help="space-separated list of t values at which to print structures \n{t| 0 <= t <= 1}"
)

args = interpolate_parser.parse_args()

if not any([args.specific_ts, args.print_E, args.print_min, args.print_max, args.n_struc]):
    interpolate_parser.print_help(sys.stderr)
    raise RuntimeError(
        "one of the following flags should be used: %s" % ", ".join(instruction_args)
    )

if args.use_incomplete and args.print_E:
    warn(
        "Any unfinished geometry optimizations will be used to interpolate energies.\n" +
        "Results may be nonsense"
    )

#take ".xyz" off the outfile b/c aarontools adds it back
outfile = args.outfile[0]

#list of usable input geometries
geom_list = []
nrg_list = []

#read each input file and see if it has normal modes
for f in args.infiles:
    fr = FileReader(f, just_geom=False)
    if "finished" not in fr.other:
        #finished was never set (e.g. input was an xyz file), but we"ll use it anyways
        if args.use_incomplete:
            geom = Geometry(fr)
            geom_list.append(geom)
        else:
            warn("not using %s because it is not marked as finished" % f)
    else:
        if fr.other["finished"] or args.use_incomplete:
            #finished is False, but we"ll use it anyways
            geom = Geometry(fr)
            nrg_list.append(fr.other["energy"])
            geom_list.append(geom)
        else:
            warn("not using %s because it is not marked as finished" % f)

if len(geom_list) <= 1:
    warn(
        "nothing to interpolate: %i usable input structure%s" % (
            len(geom_list), (1 - len(geom_list)) * "s"
        )
    )
    warn("use the -u option to include structures without an associated energy")
    sys.exit(0)

ref_geom = geom_list[0]

if len(nrg_list) < len(geom_list):
    nrg_list = np.zeros(len(geom_list))

#align all input geometries to the reference
for geom, nrg in zip(geom_list, nrg_list):
    centroid = geom.COM(mass_weight=True)
    geom.coord_shift(vector=-1 * centroid)
    geom.RMSD(ref=ref_geom, align=True)

#interpolate between the structures
pathway = Pathway(
    ref_geom,
    np.array([geom.coords for geom in geom_list]),
    other_vars={"energy": nrg_list}
)
s_max, r_max = Pathway.t_to_s(1, pathway.region_length)

#header for writing energies
nrg_out = "t\tE\tdE/dt\n"
#list of geometries to print
write_geoms = []

if args.print_max or args.print_min:
    #to find minima and maxima, find where derivative crosses 0 and
    #sign of derivative at neighboring points
    max_n_min_ts = []
    ts = np.linspace(0, 1, num=10001)
    dt = ts[1] - ts[0]
    for t in ts:
        dnrg_dt = pathway.dvar_func_dt["energy"](t) * pathway.dvar_func_dt["energy"](t + dt)
        if dnrg_dt <= 0 and pathway.dvar_func_dt["energy"](t) > 0 and args.print_max:
            max_n_min_ts.append(t)
        elif dnrg_dt <= 0 and pathway.dvar_func_dt["energy"](t) < 0 and args.print_min:
            max_n_min_ts.append(t)

    for i, t in enumerate(max_n_min_ts):
        nrg = pathway.var_func["energy"](t)
        if args.print_E:
            d_nrg = pathway.dvar_func_dt["energy"](t)
            nrg_out += "%f\t%f\t%f\n" % (t, nrg, d_nrg)
        else:
            geom = pathway.geom_func(t)
            comment = "E(%f) = %f" % (t, nrg)
            geom.comment = comment
            write_geoms.append(geom)

if args.specific_ts:
    #print structures for specified values of t
    for i, t in enumerate(args.specific_ts):
        if args.print_E:
            nrg = pathway.var_func["energy"](t)
            d_nrg = pathway.dvar_func_dt["energy"](t)
            nrg_out += "%f\t%f\t%f\n" % (t, nrg, d_nrg)
        else:
            geom = pathway.geom_func(t)
            nrg = pathway.var_func["energy"](t)
            comment = "E(%f) = %f" % (t, nrg)
            geom.comment = comment
            write_geoms.append(geom)

if args.print_E:
    if args.n_struc:
        ss = np.linspace(0, s_max, num=args.n_struc[0])
        for s in ss:
            t = Pathway.s_to_t(s, pathway.region_length)
            nrg = pathway.var_func["energy"](t)
            d_nrg = pathway.dvar_func_dt["energy"](t)
            nrg_out += "%f\t%f\t%f\n" % (t, nrg, d_nrg)

    if outfile is not None:
        with open(outfile, "w") as f:
            f.write(nrg_out.rstrip())
    else:
        print(nrg_out.rstrip())

else:
    if args.n_struc:
        w = width(args.n_struc[0])
        fmt = "%0" + "%i" % w + "i"

        ts = np.linspace(0, 1, num=args.n_struc[0])
        for i, t in enumerate(ts):
            geom = pathway.geom_func(t)
            nrg = pathway.var_func["energy"](t)
            comment = "E(%f) = %f" % (t, nrg)
            geom.comment = comment
            write_geoms.append(geom)

if len(write_geoms) > 0:
    w = width(len(write_geoms))
    fmt = "%0" + "%i" % w + "i"

if outfile is None:
    outfile = "traj-$i.xyz"

for i, geom in enumerate(write_geoms):
    my_outfile = outfile.replace("$i", fmt % i)
    if my_outfile == outfile:
        #if there"s no $i, we are writing all the structures to the same file
        geom.write(append=True, outfile=my_outfile)
    else:
        geom.write(append=False, outfile=my_outfile)



#! /usr/bin/env python3
import os
import random
import unittest

import numpy as np
from AaronTools.atoms import Atom, BondOrder
from AaronTools.geometry import Geometry
from AaronTools.test import TestWithTimer, prefix


class TestBondOrder(TestWithTimer):
    benzene_NO2_Cl = os.path.join(
        prefix, "test_files", "benzene_1-NO2_4-Cl.xyz"
    )

    def test_get(self):
        mol = Geometry(TestBondOrder.benzene_NO2_Cl)

        bond_data = BondOrder()
        for i, a in enumerate(mol.atoms):
            for c in mol.atoms[i + 1 :]:
                if c not in a.connected:
                    continue
                bo = bond_data.get(a, c)
                if a.element == "C" and c.element == "C":
                    # CC bonds should all be aromatic
                    self.assertEqual(bo, 1.5)
                elif (a.element == "N" and c.element == "O") or (
                    a.element == "O" and c.element == "N"
                ):
                    # ON bonds are 1.5
                    self.assertEqual(bo, 1.5)
                else:
                    # all other bonds are single
                    self.assertEqual(bo, 1)
        return


class TestAtoms(TestWithTimer):
    # Test constants
    benzene_NO2_Cl = os.path.join(
        prefix, "test_files", "benzene_1-NO2_4-Cl.xyz"
    )
    pentane = os.path.join(prefix, "test_files", "pentane.xyz")
    hexanol_diamine = os.path.join(prefix, "test_files", "6a2e5am1hex.xyz")

    # helper functions
    def read_xyz(self, fname):
        """just here to limit reliance on FileIO or Geometry"""
        mol = []
        with open(fname) as f:
            # split xyz file into info matrix
            tmp = f.read()
            tmp = tmp.split("\n")
            for line in tmp:
                line = line.strip().split()
                if len(line):
                    mol += [line]
            # remove atom number and comment line
            mol = mol[2:]
        atoms = []
        for i, a in enumerate(mol):
            a = Atom(element=a[0], coords=a[1:4], name=(i + 1), tags=i)
            atoms += [a]
        return atoms

    # test cases
    def test_store_atoms(self):
        atoms_read = self.read_xyz(TestAtoms.benzene_NO2_Cl)
        atoms_manual = []
        mol = []
        with open(TestAtoms.benzene_NO2_Cl) as f:
            # split xyz file into info matrix
            tmp = f.read()
            tmp = tmp.split("\n")
            for line in tmp:
                line = line.strip().split()
                if len(line):
                    mol += [line]
            # remove atom number and comment line
            mol = mol[2:]
        for i, line in enumerate(mol):
            a = Atom()
            a.element = line[0]
            a.coords = np.array(line[1:4], dtype=float)
            a.flag = False
            a.name = str(i + 1)
            a.tags = set([i])
            atoms_manual += [a]

        # test manual and **kwargs assignment match
        for a, b in zip(atoms_read, atoms_manual):
            self.assertEqual(a.element, b.element)
            self.assertTrue(np.linalg.norm(a.coords - b.coords) < 10 ** (-8))
            self.assertEqual(a.name, b.name)
            self.assertEqual(a.tags, b.tags)
            # should have automatic lookup of _radii
            self.assertTrue(a._radii is not None)

        # transform back into xyz format for verification
        atom_matrix = []
        for a in atoms_read:
            tmp = [str(a.element)]
            for c in a.coords:
                tmp += ["{:7.5f}".format(c)]
            atom_matrix += [tmp]
        for a, b in zip(mol, atom_matrix):
            a = " ".join(a)
            b = " ".join(b)
            self.assertEqual(a, b)

        return

    # utilities
    def test_float(self):
        atom_args = [
            {"element": "H", "coords": [0, 0, 0], "name": "1"},
            {"element": "H", "coords": [0, 0, 0], "name": "1.1"},
            {"element": "H", "coords": [0, 0, 0], "name": "1.1.1"},
        ]
        atom_floats = [1.0, 1.1, 1.1]
        for kwarg, ref in zip(atom_args, atom_floats):
            atom = Atom(**kwarg)
            self.assertTrue(float(atom) == ref)

    def test_less_than(self):
        atoms = Geometry(TestAtoms.benzene_NO2_Cl)
        atoms = atoms.atoms
        for atom in atoms:
            atom._rank = None
        # atom[11] has 3 connected atoms
        # atom[6] has 1 connected atom
        self.assertTrue(atoms[11] < atoms[6])
        # atom[12] is O
        # atom[10] is Cl
        self.assertTrue(atoms[12] < atoms[10])
        # atom[3] has 3 connected atoms: C, C, Cl
        # atom[2] has 3 connected atoms: C, C, H
        self.assertTrue(atoms[3] < atoms[2])

    def test_repr(self):
        atom = Atom("H", [0, 0, 0], True, "1", ["test"])
        s = repr(atom)
        self.assertEqual(
            s, "  H     0.00000000    0.00000000    0.00000000  -1    1"
        )

    def test_set_radii(self):
        atom = Atom("H", [0, 0, 0])
        self.assertEqual(atom._radii, 0.32)
        atom = Atom("X", [0, 0, 0])  # ghost atom
        self.assertEqual(atom._radii, 0.0)
        with self.assertRaises(ValueError):
            atom = Atom("NotAnElement", [0, 0, 0])
        atom = Atom()
        atom.element = "NotAnElement"
        with self.assertLogs(logger=Atom.LOG, level="WARNING"):
            atom._set_radii()

    def test_set_connectivity(self):
        atom = Atom("H", [0, 0, 0])
        self.assertEqual(atom._connectivity, 1)
        atom = Atom("C", [0, 0, 0])
        self.assertEqual(atom._connectivity, 4)

    def test_add_tag(self):
        mol = self.read_xyz(TestAtoms.benzene_NO2_Cl)
        for atom in mol:
            new_tags = [random.random() for i in range(10)]
            old_tags = atom.tags
            atom.add_tag(new_tags[0])
            atom.add_tag(new_tags[1], new_tags[2], new_tags[3], new_tags[4])
            atom.add_tag(new_tags[5:])
            self.assertSequenceEqual(
                sorted(old_tags.union(new_tags)), sorted(atom.tags)
            )

    def test_get_invariant(self):
        pentane = Geometry(TestAtoms.pentane)
        mol = Geometry(TestAtoms.hexanol_diamine)

        s = ""
        for a in pentane.atoms:
            if a.element == "H":
                continue
            s += a.get_invariant() + " "
        self.assertEqual(s, "10100063 20200062 20200062 20200062 10100063 ")

        s = ""
        for a in mol.atoms:
            if a.element == "H":
                continue
            s += a.get_invariant() + " "
        ans = "30300061 20200062 20200062 30300061 20200062 20200062 20200062 20200062 10100081 10100063 10100072 10100072 "
        self.assertEqual(s, ans)

    # measurement
    def test_is_connected(self):
        atoms = self.read_xyz(TestAtoms.benzene_NO2_Cl)
        conn_valid = [
            "2,6,12",
            "1,3,7",
            "2,4,8",
            "3,5,11",
            "4,6,9",
            "1,5,10",
            "2",
            "3",
            "5",
            "6",
            "4",
            "1,13,14",
            "12",
            "12",
        ]
        for i, a in enumerate(atoms):
            for j, b in enumerate(atoms[i + 1 :]):
                if a.is_connected(b):
                    a.connected.add(b)
                    b.connected.add(a)
            conn_try = sorted([int(c.name) for c in a.connected])
            conn_try = ",".join([str(c) for c in conn_try])
            self.assertEqual(conn_try, conn_valid[i])
        return

    def test_bond(self):
        a1 = Atom(element="H", coords=[0, 0, 0])
        a2 = Atom(element="H", coords=[1, 0, 0])
        self.assertTrue(np.linalg.norm(a1.bond(a2) - [1, 0, 0]) == 0)

    def test_dist(self):
        a1 = Atom(element="H", coords=[0, 0, 0])
        a2 = Atom(element="H", coords=[1, 0, 0])
        self.assertTrue(a1.dist(a2) == 1)

    def test_angle(self):
        mol = self.read_xyz(TestAtoms.benzene_NO2_Cl)
        angle = mol[11].angle(mol[12], mol[13])
        self.assertTrue(abs(np.rad2deg(angle) - 124.753) < 10 ** (-3))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import argparse
import sys

import numpy as np

from AaronTools.comp_output import CompOutput
from AaronTools.const import UNIT
from AaronTools.fileIO import FileReader
from AaronTools.finders import NotAny, AnyTransitionMetal, AnyNonTransitionMetal
from AaronTools.geometry import Geometry
from AaronTools.substituent import Substituent
from AaronTools.utils.utils import boltzmann_coefficients, glob_files

def main(argv):
    vbur_parser = argparse.ArgumentParser(
        description="calculate Boltzmann-weighted percent buried volume parameters",
        formatter_class=argparse.RawTextHelpFormatter
    )
    
    vbur_parser.add_argument(
        "infiles", metavar="input files",
        type=str,
        nargs="+",
        help="file containing coordinates and energy"
    )
    
    vbur_parser.add_argument(
        "-if", "--input-format",
        type=str,
        default=None,
        choices=["log", "out", "dat"],
        dest="input_format",
        help="file format of input"
    )
    
    vbur_parser.add_argument(
        "-o", "--output",
        type=str,
        default=False,
        required=False,
        metavar="output destination",
        dest="outfile",
        help="output destination\n" +
        "Default: stdout"
    )

    vbur_parser.add_argument(
        "-l", "--ligand-atoms",
        default=None,
        required=False,
        dest="targets",
        help="atoms to consider in calculation\nDefault: use all atoms except the center",
    )
    
    vbur_parser.add_argument(
        "-e", "--exclude-atoms",
        default=None,
        required=False,
        dest="exclude_atoms",
        help="atoms to exclude from the calculation\nDefault: exclude no ligand atoms",
    )
    
    vbur_parser.add_argument(
        "-c", "--center",
        action="append",
        default=None,
        required=False,
        dest="center",
        help="atom the sphere is centered on\n" +
        "Default: detect metal center (centroid of all metals if multiple are present)",
    )

    vbur_parser.add_argument(
        "-r", "--radius",
        default=3.5,
        type=float,
        dest="radius",
        help="radius around center\nDefault: 3.5 Ångström"
    )

    vbur_parser.add_argument(
        "-vdw", "--vdw-radii",
        type=str,
        default="bondi",
        choices=["bondi", "umn"],
        dest="radii",
        help="VDW radii to use in calculation\n" + 
        "umn: main group vdw radii from J. Phys. Chem. A 2009, 113, 19, 5806–5812\n" +
        "    (DOI: 10.1021/jp8111556)\n" + 
        "    transition metals are crystal radii from Batsanov, S.S. Van der Waals\n" +
        "    Radii of Elements. Inorganic Materials 37, 871–885 (2001).\n" +
        "    (DOI: 10.1023/A:1011625728803)\n" + 
        "bondi: radii from J. Phys. Chem. 1964, 68, 3, 441–451 (DOI: 10.1021/j100785a001)\n" +
        "Default: bondi"
    )

    vbur_parser.add_argument(
        "-s", "--scale",
        default=1.17,
        type=float,
        dest="scale",
        help="scale VDW radii by this amount\nDefault: 1.17",
    )
    
    vbur_parser.add_argument(
        "-t", "--temperature",
        type=str,
        default=298.15,
        required=False,
        dest="temperature",
        help="temperature in K\nDefault: 298.15"
    )
    
    vbur_parser.add_argument(
        "-f", "--frequency",
        action="store_true",
        default=False,
        required=False,
        dest="frequency",
        help="input files are frequency job output files\n"
        "additional average values will be calculated for ZPE, H, G, etc."
    )
    
    vbur_parser.add_argument(
        "-w0",
        "--frequency-cutoff",
        type=float,
        default=100.0,
        required=False,
        dest="w0",
        help="cutoff frequency for quasi free energy corrections (1/cm)\n" +
        "Default: 100 cm^-1"
    )
    
    vbur_parser.add_argument(
        "-v", "--verbose",
        action="store_true",
        default=False,
        required=False,
        dest="verbose",
        help="also print population"
    )

    vbur_parser.add_argument(
        "-m", "--method",
        default="Lebedev",
        type=lambda x: x.capitalize() if x.lower() == "lebedev" else x.upper(),
        choices=["MC", "Lebedev"],
        dest="method",
        help="integration method - Monte-Carlo (MC) or Lebedev quadrature (Lebedev)\nDefault: Lebedev"
    )
    
    grid_options = vbur_parser.add_argument_group("Lebedev integration options")
    grid_options.add_argument(
        "-rp", "--radial-points",
        type=int,
        default=20,
        choices=[20, 32, 64, 75, 99, 127],
        dest="rpoints",
        help="number of radial shells for Gauss-Legendre integration\n" +
        "of the radial component\n" +
        "lower values are faster, but at the cost of accuracy\nDefault: 20"
    )
    
    grid_options.add_argument(
        "-ap", "--angular-points",
        type=int,
        default=1454,
        choices=[110, 194, 302, 590, 974, 1454, 2030, 2702, 5810],
        dest="apoints",
        help="number of angular points for Lebedev integration\n" +
        "lower values are faster, but at the cost of accuracy\nDefault: 1454"
    )
    
    mc_options = vbur_parser.add_argument_group("Monte-Carlo integration options")
    mc_options.add_argument(
        "-i", "--minimum-iterations",
        type=int,
        default=25,
        metavar="ITERATIONS",
        dest="min_iter",
        help="minimum iterations - each is a batch of 3000 points\n" +
        "MC will continue after this until convergence criteria are met\n" +
        "Default: 25",
    )

    args = vbur_parser.parse_args(args=argv)
    
    targets = None
    if args.exclude_atoms and not args.targets:
        targets = (NotAny(args.exclude_atoms))
    elif args.exclude_atoms and args.targets:
        targets = (NotAny(args.exclude_atoms), args.targets)
    else:
        targets = NotAny(args.center)
    
    geoms = []
    energies = {"E":[]}
    if args.frequency:
        energies["E+ZPE"] = []
        energies["H(RRHO)"] = []
        energies["G(RRHO)"] = []
        energies["G(Quasi-RRHO)"] = []
        energies["G(Quasi-Harmonic)"] = []
    
    for infile in glob_files(args.infiles, parser=vbur_parser):
        if args.input_format is not None:
            fr = FileReader((infile, args.input_format, infile), just_geom=False)
        else:
            fr = FileReader(infile, just_geom=False)
    
        geom = Geometry(fr)
        geoms.append(geom)
        
        nrg = fr.other["energy"]
        energies["E"].append(nrg)
        if args.frequency:
            co = CompOutput(fr)
            dE, dH, entropy = co.therm_corr(temperature=args.temperature)
            rrho_dG = co.calc_G_corr(v0=0, temperature=args.temperature, method="RRHO")
            qrrho_dG = co.calc_G_corr(v0=args.w0, temperature=args.temperature, method="QRRHO")
            qharm_dG = co.calc_G_corr(v0=args.w0, temperature=args.temperature, method="QHARM")
            energies["E+ZPE"].append(nrg + co.ZPVE)
            energies["H(RRHO)"].append(nrg + dH)
            energies["G(RRHO)"].append(nrg + rrho_dG)
            energies["G(Quasi-RRHO)"].append(nrg + qrrho_dG)
            energies["G(Quasi-Harmonic)"].append(nrg + qharm_dG)
    
    s = ""
    for nrg_type in energies:
        energies_arr = np.array(energies[nrg_type])
        energies_arr *= UNIT.HART_TO_KCAL
        if args.verbose and nrg_type == "E":
            s += "\t".join(["%Vbur", "file"])
            s += "\n"
            for f, geom in zip(args.infiles, geoms):
                data = geom.percent_buried_volume(
                    targets=targets,
                    center=args.center,
                    radius=args.radius,
                    radii=args.radii,
                    scale=args.scale,
                    method=args.method,
                    rpoints=args.rpoints,
                    apoints=args.apoints,
                    min_iter=args.min_iter,
                )
                s += "%.1f%%\t%s\n" % (data, f)
            s += "\n"

        s += "weighted using %s:\n" % nrg_type
        data = Geometry.weighted_percent_buried_volume(
            geoms,
            energies_arr,
            args.temperature,
            targets=targets,
            center=args.center,
            radius=args.radius,
            radii=args.radii,
            scale=args.scale,
            method=args.method,
            rpoints=args.rpoints,
            apoints=args.apoints,
            min_iter=args.min_iter,
        )

        if args.verbose:
            coeff = boltzmann_coefficients(energies_arr, args.temperature)
            coeff /= sum(coeff)
            coeff *= 100
            for f, c, e in zip(args.infiles, coeff, energies_arr):
                s += "%s  %.1f%% (%.1f kcal/mol)\n" % (f, c, e - min(energies_arr))

        s += "%%Vbur: %.2f\n\n" % data
    
    if not args.outfile:
        print(s)
    else:
        with open(args.outfile, "w") as f:
            f.write(s)

if __name__ == "__main__":
    main(None)



#!/usr/bin/env python3
import os
import unittest

from AaronTools import addlogger
from AaronTools.config import Config
from AaronTools.test import TestWithTimer, prefix
from AaronTools.theory import Theory


@addlogger
class TestTheory(TestWithTimer):
    LOG = None
    USER_SPECIFIC = ["metadata", "infile", (".*", ".*_dir")]

    def test_xtb_cmdline(self):
        config = Config(
            os.path.join(prefix, "test_files", "xtb_config.ini"),
            skip_user_default=True,
        )
        config = config.for_change("OH_Cl")
        structure = config.get_template()[0][0]
        structure = config.make_changes(structure)

        this_config = config.for_step(1)
        theory = this_config.get_theory(structure)
        ref = {
            "--opt": None,
            "--uhf": 1,
            "--etemp": "400",
            "--alpb": "acetone",
            "--restart": None,
            "-P": "2",
        }
        test = theory.get_xtb_cmdline(this_config)
        self.assertDictEqual(test, ref)

        this_config = config.for_step(2)
        theory = this_config.get_theory(structure)
        ref = {
            "--optts": None,
            "--uhf": 1,
            "--etemp": "400",
            "--gbsa": "acetone",
        }
        test = theory.get_xtb_cmdline(this_config)
        self.assertDictEqual(test, ref)

    def test_copy_equal(self):
        ref = Theory(
            method="B3LYP",
            basis="def2-SVP",
            job_type="sp",
            grid="(99, 590)",
        )
        
        test = ref.copy()
        self.assertEqual(test, ref)

def suite():
    suite = unittest.TestSuite()
    suite.addTest(TestTheory("test_xtb_cmdline"))
    suite.addTest(TestTheory("test_copy_equal"))
    return suite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/env python3

import itertools
import os
import re
from glob import glob

from AaronTools.const import AARONLIB, AARONTOOLS
from AaronTools.fileIO import FileReader, read_types
from AaronTools.geometry import Geometry


class Ring(Geometry):
    """
    Attributes:
        name
        atoms
        end
    """

    AARON_LIBS = os.path.join(AARONLIB, "Rings")
    BUILTIN = os.path.join(AARONTOOLS, "Rings")

    def __init__(self, frag, name=None, end=None):
        """
        frag is either a file sub, a geometry, or an atom list
        name is a name
        end is a list of atoms that defines which part of the ring is not part of the fragment
        """

        super().__init__()
        self.end = end
        if isinstance(frag, (Geometry, list)):
            # we can create ring object from a geometry
            if isinstance(frag, Ring):
                self.name = name if name else frag.name
                self.end = end if end else frag.end
            elif isinstance(frag, Geometry):
                self.name = name if name else frag.name
                self.end = end if end else None
            else:
                self.name = name

            try:
                self.atoms = frag.atoms
            except AttributeError:
                self.atoms = frag

        else:  # or we can create from file
            # find ring xyz file
            fring = None
            for lib in [Ring.AARON_LIBS, Ring.BUILTIN]:
                if not os.path.exists(lib):
                    continue
                for f in os.listdir(lib):
                    name, ext = os.path.splitext(f)
                    if not any(".%s" % x == ext for x in read_types):
                        continue
                    match = frag == name
                    if match:
                        fring = os.path.join(lib, f)
                        break
                
                if fring:
                    break
            # or assume we were given a file name instead
            if not fring and ".xyz" in frag:
                fring = frag
                frag = os.path.basename(frag).rstrip(".xyz")

            if fring is None:
                raise RuntimeError("ring name not recognized: %s" % frag)

            # load in atom info
            from_file = FileReader(fring)
            self.name = frag
            self.comment = from_file.comment
            self.atoms = from_file.atoms
            self.refresh_connected()

            end_info = re.search("E:(\d+)", self.comment)
            if end_info is not None:
                self.end = [
                    self.find(end)[0]
                    for end in re.findall("\d+", self.comment)
                ]
            else:
                self.end = None

    @classmethod
    def from_string(cls, name, end_length, end_atom=None, form="smiles"):
        """create ring fragment from string
        name        str         identifier for ring
        end_length  int         number of atoms in ring end
        end_atom    identifiers identifier for ring end
        form        str         type of identifier (smiles, iupac)
        """

        ring = Geometry.from_string(name, form)
        if end_atom is not None and end_length is not None:
            ring = cls(ring)
            end_atom = ring.find(end_atom)[0]
            ring.find_end(end_length, end_atom)
            return ring

        elif end_length is not None:
            ring = cls(ring)
            ring.find_end(end_length)
            return ring

        else:
            return cls(ring, name=name)

    @classmethod
    def list(cls, include_ext=False):
        names = []
        for lib in [cls.AARON_LIBS, cls.BUILTIN]:
            if not os.path.exists(lib):
                continue
            for f in os.listdir(lib):
                name, ext = os.path.splitext(os.path.basename(f))
                if not any(".%s" % x == ext for x in read_types):
                    continue
                
                if name in names:
                    continue
                
                if include_ext:
                    names.append(name + ext)
                else:
                    names.append(name)

        return names

    def copy(self):
        dup = super().copy()
        dup.end = dup.find([atom.name for atom in self.end])
        return Ring(dup, end=dup.end)

    def find_end(self, path_length, start=[]):
        """finds a path around self that is path_length long and starts with start"""
        def linearly_connected(atom_list):
            """returns true if every atom in atom_list is connected to another atom in
            the list without backtracking"""
            # start shouldn't be end
            if atom_list[0] == atom_list[-1]:
                return False

            # first and second atoms should be bonded
            elif atom_list[0] not in atom_list[1].connected:
                return False

            # last and second to last atoms should be bonded
            elif atom_list[-1] not in atom_list[-2].connected:
                return False

            # all other atoms should be conneced to exactly 2 atoms
            elif any(
                [
                    sum([atom1 in atom2.connected for atom2 in atom_list]) != 2
                    for atom1 in atom_list[1:-1]
                ]
            ):
                return False

            # first two atoms should only be connected to one atom unless they are connected to each other
            elif (
                sum(
                    [
                        sum([atom_list[0] in atom.connected])
                        + sum([atom_list[-1] in atom.connected])
                        for atom in atom_list
                    ]
                )
                > 2
                and atom_list[0] not in atom_list[-1].connected
            ):

                return False

            else:
                return True

        self.end = None

        if start:
            start_atoms = self.find(start)
        else:
            start_atoms = []

        usable_atoms = []
        for atom in self.atoms:
            if atom not in start_atoms:
                if hasattr(atom, "_connectivity"):
                    if atom._connectivity > 1:
                        usable_atoms.append(atom)
                else:
                    usable_atoms.append(atom)

        for path in itertools.permutations(
            usable_atoms, path_length - len(start_atoms)
        ):
            full_path = start_atoms + list(path)
            if linearly_connected(full_path) or path_length == 1:
                self.end = list(full_path)
                break

        if self.end is None:
            raise LookupError(
                "unable to find %i long path starting with %s around %s"
                % (path_length, start, self.name)
            )

        else:
            self.comment = "E:" + ",".join(
                [str(self.atoms.index(a) + 1) for a in self.end]
            )



#!/usr/bin/env python3
import json
import os
import unittest
from copy import copy

import AaronTools
import numpy as np
from AaronTools.atoms import Atom
from AaronTools.component import Component
from AaronTools.fileIO import FileReader
from AaronTools.geometry import Geometry
from AaronTools.ring import Ring
from AaronTools.substituent import Substituent
from AaronTools.test import TestWithTimer, prefix, rmsd_tol, validate


def is_close(a, b, tol=10 ** -8, debug=False):
    n = None
    if isinstance(a, list):
        a = np.array(a)
    if isinstance(b, list):
        b = np.array(b)
    if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
        n = np.linalg.norm(a - b)
    elif isinstance(a, np.ndarray):
        n = np.linalg.norm(a) - b
    elif isinstance(b, np.ndarray):
        n = a - np.linalg.norm(b)
    else:
        n = a - b
    if debug:
        print(abs(n))
    return abs(n) < tol


class TestGeometry(TestWithTimer):
    benz_NO2_Cl = os.path.join(prefix, "test_files", "benzene_1-NO2_4-Cl.xyz")
    benz_NO2_Cl_conn = [
        "2,6,12",
        "1,3,7",
        "2,4,8",
        "3,5,11",
        "4,6,9",
        "1,5,10",
        "2",
        "3",
        "5",
        "6",
        "4",
        "1,13,14",
        "12",
        "12",
    ]
    benz_NO2_Cl_conn = [i.split(",") for i in benz_NO2_Cl_conn]
    benzene = os.path.join(prefix, "test_files", "benzene.xyz")
    pyridine = os.path.join(prefix, "test_files", "pyridine.xyz")
    pentane = os.path.join(prefix, "test_files", "pentane.xyz")
    naphthalene = os.path.join(prefix, "ref_files", "naphthalene.xyz")
    tetrahydronaphthalene = os.path.join(
        prefix, "ref_files", "tetrahydronaphthalene.xyz"
    )
    pyrene = os.path.join(prefix, "ref_files", "pyrene.xyz")
    benz_Cl = os.path.join(prefix, "test_files", "benzene_4-Cl.xyz")
    benz_OH_Cl = os.path.join(prefix, "test_files", "benzene_1-OH_4-Cl.xyz")
    benz_Ph_Cl = os.path.join(prefix, "test_files", "benzene_1-Ph_4-Cl.xyz")
    Et_NO2 = os.path.join(prefix, "test_files", "Et_1-NO2.xyz")
    cat = os.path.join(prefix, "test_files", "catalysts", "tm_single-lig.xyz")

    tm_simple = os.path.join(
        prefix, "test_files", "catalysts/tm_single-lig.xyz"
    )
    tm_multi = os.path.join(
        prefix, "test_files", "catalysts", "tm_multi-lig.xyz"
    )
    org_1 = os.path.join(prefix, "test_files", "catalysts", "org_1.xyz")
    org_tri = os.path.join(prefix, "test_files", "catalysts", "org_tri.xyz")
    ptco4 = os.path.join(prefix, "test_files", "ptco4.xyz")
    catalysts = [tm_simple, tm_multi, org_1, org_tri, ptco4]

    monodentate = os.path.join(prefix, "test_files", "ligands", "ACN.xyz")
    bidentate = os.path.join(prefix, "test_files", "ligands", "S-tBu-BOX.xyz")
    tridentate = os.path.join(
        prefix, "test_files", "ligands", "squaramide.xyz"
    )

    lig_1 = os.path.join(prefix, "test_files", "lig_1.xyz")
    lig_2 = os.path.join(prefix, "test_files", "lig_2.xyz")

    def test_init(self):
        ref = FileReader(TestGeometry.benz_NO2_Cl)
        conn_valid = TestGeometry.benz_NO2_Cl_conn
        rank_valid = [8, 7, 5, 4, 5, 7, 3, 1, 1, 3, 0, 6, 2, 2]

        for i, a in enumerate(ref.atoms):
            a.connected = set(conn_valid[i])
            a._rank = rank_valid[i]

        mol = Geometry(TestGeometry.benz_NO2_Cl)

        # name
        self.assertEqual(mol.name, ref.name)
        # comment
        self.assertEqual(mol.comment, ref.comment)
        # atoms
        for a, b in zip(ref.atoms, mol.atoms):
            for attr in a.__dict__:
                if attr == "file_type":
                    continue
                if attr == "connected":
                    tmpa = [int(c) for c in a.connected]
                    tmpb = [int(c.name) for c in b.connected]
                    self.assertSequenceEqual(sorted(tmpa), sorted(tmpb))
                    continue
                if attr == "_rank":
                    continue
                if attr == "constraint":
                    # this is checked in test_parse_comment
                    continue
                try:
                    self.assertEqual(a.__dict__[attr], b.__dict__[attr])
                except ValueError:
                    self.assertSequenceEqual(
                        sorted(a.__dict__[attr]), sorted(b.__dict__[attr])
                    )

        # test blank
        blank = Geometry()
        self.assertEqual(blank.name, "")
        self.assertEqual(blank.comment, "")
        self.assertEqual(blank.atoms, [])

    def test_attribute_access(self):
        mol = Geometry(TestGeometry.benzene)
        # stack coords
        coords = mol._stack_coords()
        self.assertEqual(coords.shape, (12, 3))
        # elements
        elements = mol.elements
        self.assertEqual(len(elements), 12)
        self.assertEqual(elements[0], "C")
        # coords
        coords = mol.coords
        self.assertEqual(coords.shape, (12, 3))

    # utilities
    def test_equality(self):
        mol = Geometry(TestGeometry.benz_NO2_Cl)
        benz = Geometry(TestGeometry.benzene)
        # same object should be equal
        self.assertEqual(mol, mol)
        # copy should be equal
        self.assertEqual(mol, mol.copy())
        # different molecules should be unequal
        self.assertNotEqual(mol, benz)

    def test_add_sub_iter_len(self):
        ref = Geometry(TestGeometry.benz_OH_Cl)
        mol = Geometry(TestGeometry.benzene)
        mol.debug = True
        OH = ref.find(["12", "13"])
        Cl = ref.find(["11"])
        mol -= mol.find(["11", "12"])
        mol += Cl
        mol += OH
        mol.refresh_connected()
        mol.refresh_ranks()
        self.assertTrue(validate(mol, ref, thresh="loose"))

    def test_find_atom(self):
        geom = Geometry(TestGeometry.benz_NO2_Cl)
        geom.atoms[0].add_tag("find_me")

        # find a specific atom
        a = geom.find(geom.atoms[0])
        self.assertEqual(a, geom.atoms[0:1])

        # find by tag
        a = geom.find("find_me")
        self.assertEqual(a, geom.atoms[0:1])

        # find by name
        a = geom.find("1")
        self.assertEqual(a, geom.atoms[0:1])

        # find by list-style name
        a = geom.find("1,2-5,12")
        self.assertSequenceEqual(
            a, geom.atoms[0:1] + geom.atoms[1:5] + geom.atoms[11:12]
        )

        # find using tag and name
        a = geom.find("find_me", "1")
        self.assertEqual(a, geom.atoms[0:1])

        b = geom.find(["find_me", "2"])
        self.assertEqual(b, geom.atoms[0:2])

        c = geom.find(["1", "2"], "find_me")
        self.assertEqual(c, geom.atoms[0:1])

        d = geom.find(["2", "3"], "find_me")
        self.assertEqual(d, [])

        # find all Carbons
        a = geom.find("C")
        b = []
        for i in geom.atoms:
            if i.element == "C":
                b += [i]
        self.assertSequenceEqual(a, b)

        # raise error when atom not found
        self.assertRaises(LookupError, geom.find, "definitely not in here")

    def test_refresh_connected(self):
        # refresh_connected should be run upon creation
        mol = Geometry(TestGeometry.benz_NO2_Cl)
        conn_valid = TestGeometry.benz_NO2_Cl_conn
        for a, b in zip(mol.atoms, conn_valid):
            tmpa = [int(c.name) for c in a.connected]
            tmpb = [int(c) for c in b]
            self.assertSequenceEqual(sorted(tmpa), sorted(tmpb))

        # old connectivity shouldn't remain in the set
        mol.atoms[0].connected.add(Atom())
        old = mol.atoms[0].connected
        mol.refresh_connected()
        self.assertTrue(len(old) - len(mol.atoms[0].connected) == 1)

    def test_canonical_rank(self):
        pentane = Geometry(os.path.join(prefix, "test_files", "pentane.xyz"))
        pentane_rank = [1, 3, 4, 2, 0]
        test_rank = pentane.canonical_rank(heavy_only=True)
        self.assertSequenceEqual(test_rank, pentane_rank)

        mol = Geometry(os.path.join(prefix, "test_files", "6a2e5am1hex.xyz"))
        mol_rank = [8, 6, 5, 7, 2, 1, 4, 3, 11, 0, 10, 9]
        test_rank = mol.canonical_rank(heavy_only=True, invariant=False)
        self.assertSequenceEqual(test_rank, mol_rank)

    def test_flag(self):
        geom = Geometry(TestGeometry.benz_NO2_Cl)

        # freeze all
        test = geom.copy()
        test.freeze()
        for a in test.atoms:
            self.assertTrue(a.flag)

        # freeze some
        test = geom.copy()
        test.freeze("C")
        for a in test.atoms:
            if a.element == "C":
                self.assertTrue(a.flag)
            else:
                self.assertFalse(a.flag)

        geom.freeze()
        # relax all
        test = geom.copy()
        test.relax()
        for a in test.atoms:
            self.assertFalse(a.flag)

        # relax some
        test = geom.copy()
        test.relax("C")
        for a in test.atoms:
            if a.element == "C":
                self.assertFalse(a.flag)
            else:
                self.assertTrue(a.flag)

    def test_update_geometry(self):
        test = Geometry(TestGeometry.benz_NO2_Cl)

        # using coordinate matrix to update
        ref = test.copy()
        ref.coord_shift([-10, 0, 0])
        tmp = test._stack_coords()
        for t in tmp:
            tmp = tmp - np.array([-10, 0, 0])
        test.update_geometry(tmp)
        self.assertTrue(validate(test, ref))

        # using file
        ref.coord_shift([10, 0, 0])
        test.update_geometry(TestGeometry.benz_NO2_Cl)
        self.assertTrue(validate(test, ref))

    def test_near(self):
        def compare(test, ref):
            self.assertTrue(len(test) == len(set(test)))
            test = set(test)
            for a in ref:
                self.assertTrue(a in test)
                test.discard(a)
            self.assertTrue(len(test) == 0)

        geom = Geometry(self.benz_NO2_Cl)

        # atoms within 3A of origin
        test = geom.get_near([0, 0, 0], 3)
        ref = geom.find(["2", "3", "4", "5", "6", "8", "9", "11"])
        compare(test, ref)
        # atoms within 1A of x-axis
        test = geom.get_near(["*", 0, 0], 1)
        ref = geom.find(["3", "4"])
        compare(test, ref)
        # atoms within 0.5A of xy-plane
        test = geom.get_near(["*", "*", 0], 0.5)
        ref = geom.find(["1", "4", "11", "12", "13", "14"])
        compare(test, ref)

        # atoms within 2A of atom 1
        test = geom.get_near(geom.atoms[0], 2)
        ref = geom.find(["2", "6", "12"])
        compare(test, ref)
        # ...including atom 1
        test = geom.get_near(geom.atoms[0], 2, include_ref=True)
        ref = geom.find(["1", "2", "6", "12"])
        compare(test, ref)
        # atoms within 2 bonds of atom 1
        test = geom.get_near(geom.atoms[0], 2, by_bond=True)
        ref = geom.find(["2", "6", "12", "3", "7", "5", "10", "13", "14"])
        compare(test, ref)
        # atoms within 1 bond of atoms 1 and 2
        test = geom.get_near(geom.atoms[:2], 1, by_bond=True)
        ref = geom.find(["6", "12", "3", "7"])
        compare(test, ref)
        # ...including starting atoms
        test = geom.get_near(geom.atoms[:2], 1, by_bond=True, include_ref=True)
        ref = geom.find(["6", "12", "3", "7", "2", "1"])
        compare(test, ref)

    def test_compare_connectivity(self):
        geom = Geometry(TestGeometry.cat)
        ref = Geometry(TestGeometry.cat)

        # no formed/broken
        broken, formed = geom.compare_connectivity(ref)
        self.assertTrue(len(broken) == 0)
        self.assertTrue(len(formed) == 0)
        # broken
        geom.change_distance("10", "15", dist=1, adjust=True)
        geom.refresh_connected()
        broken, formed = geom.compare_connectivity(ref)
        self.assertTrue(len(broken) == 1)
        self.assertTrue(len(formed) == 0)
        self.assertSetEqual(broken, set([("10", "15")]))
        # formed
        ref.change_distance("10", "15", dist=1, adjust=True)
        ref.refresh_connected()
        geom.change_distance("10", "15", dist=-1, adjust=True)
        geom.refresh_connected()
        broken, formed = geom.compare_connectivity(ref)
        self.assertTrue(len(broken) == 0)
        self.assertTrue(len(formed) == 1)
        self.assertSetEqual(formed, set([("10", "15")]))
        # broken and formed
        geom.change_distance("20", "29", dist=1, adjust=True)
        geom.refresh_connected()
        broken, formed = geom.compare_connectivity(ref)
        self.assertTrue(len(broken) == 1)
        self.assertTrue(len(formed) == 1)
        self.assertSetEqual(broken, set([("20", "29")]))
        self.assertSetEqual(formed, set([("10", "15")]))

    def test_examine_constraints(self):
        cat = Geometry(TestGeometry.cat)
        rv = cat.examine_constraints()
        self.assertSequenceEqual(rv, [])

        cat.change_distance(cat.atoms[0], cat.atoms[1], dist=0.5, adjust=True)
        rv = cat.examine_constraints()
        self.assertSequenceEqual(rv, [(0, 1, -1)])

        cat.change_distance(cat.atoms[0], cat.atoms[1], dist=-1.0, adjust=True)
        rv = cat.examine_constraints()
        self.assertSequenceEqual(rv, [(0, 1, 1)])

        cat.change_distance(cat.atoms[1], cat.atoms[2], dist=0.5, adjust=True)
        rv = cat.examine_constraints()
        self.assertSequenceEqual(rv, [(0, 1, 1), (1, 2, -1)])

    def test_detect_components(self):
        test = {}
        for cat in TestGeometry.catalysts:
            cat = Geometry(cat)
            cat.detect_components()
            for comp in cat.components:
                test[os.path.basename(comp.name)] = sorted(
                    [int(float(c)) for c in comp]
                )

        with open(
            os.path.join(prefix, "ref_files", "detect_components.json")
        ) as f:
            ref = json.load(f)
        self.assertDictEqual(test, ref)

    def test_fix_comment(self):
        cat = Geometry(TestGeometry.tm_simple)
        cat.fix_comment()
        self.assertEqual(
            cat.comment,
            "C:1 K:2,3;14;39,40 F:1-2;1-3;1-14;2-3;2-14 L:2-13;14-34;35-93",
        )
        cat.substitute("Me", "4")
        self.assertEqual(
            cat.comment,
            "C:1 K:2,3;14;39,40 F:1-2;1-3;1-14;2-3;2-14 L:2-13;14-34;35-93",
        )

    # geometry measurement
    def test_angle(self):
        mol = Geometry(TestGeometry.benz_NO2_Cl)
        angle = mol.angle("13", "12", "14")
        self.assertTrue(is_close(np.rad2deg(angle), 124.752, 10 ** -2))

    def test_dihedral(self):
        mol = Geometry(TestGeometry.benz_NO2_Cl)
        dihedral = mol.dihedral("13", "12", "1", "6")
        self.assertTrue(is_close(np.rad2deg(dihedral), 45.023740, 10 ** -5))

    def test_COM(self):
        mol = Geometry(TestGeometry.benz_NO2_Cl)

        # all atoms
        com = mol.COM(mass_weight=True)
        self.assertTrue(
            is_close(com, [-1.41347592e00, -1.35049427e00, 9.24797359e-04])
        )

        # only carbons
        com = mol.COM(targets="C")
        self.assertTrue(
            is_close(com, [-1.27963500e00, -1.11897500e00, 8.53333333e-04])
        )
        # mass weighting COM shouldn't change anythin if they are all C
        com = mol.COM(targets="C", mass_weight=True)
        self.assertTrue(
            is_close(com, [-1.27963500e00, -1.11897500e00, 8.53333333e-04])
        )

        # only heavy atoms
        com = mol.COM(heavy_only=True, mass_weight=True)
        self.assertTrue(
            is_close(com, [-1.41699980e00, -1.35659562e00, 9.31512531e-04])
        )

    def test_RMSD(self):
        ref = Geometry(TestGeometry.benz_NO2_Cl)

        # RMSD of copied object should be 0
        test = ref.copy()
        self.assertTrue(validate(test, ref))

        # RMSD of shifted copy should be 0
        test = ref.copy()
        test.coord_shift([1, 2, 3])
        self.assertTrue(validate(test, ref))

        # RMSD of rotated copy should be 0
        test = ref.copy()
        test.rotate([1, 2, 3], 2.8)
        self.assertTrue(validate(test, ref))

        # RMSD of two different structures should not be 0
        test = Geometry(TestGeometry.pentane)
        self.assertFalse(validate(test, ref))

        # RMSD of similar molecule
        test = Geometry(TestGeometry.benzene)
        res = ref.RMSD(test, targets="C", ref_targets="C")
        self.assertTrue(res < rmsd_tol(ref))

    def test_vbur_MC(self):
        """
        tests % volume buried (MC integration)
        uses Monte Carlo integration, so it this fails, run it again
        still figuring out how reliable it is
        """

        # import cProfile
        # 
        # profile = cProfile.Profile()
        # profile.enable()

        geom = Geometry(os.path.join(prefix, "ref_files", "lig_map_3.xyz"))
        vbur = geom.percent_buried_volume(method="MC")
        if not np.isclose(vbur, 86.0, atol=0.35):
            print("V_bur =", vbur, "expected:", 86.0)
        self.assertTrue(np.isclose(vbur, 86.0, atol=0.35))

        # a few synthetic tests
        geom2 = Geometry(os.path.join(prefix, "ref_files", "vbur.xyz"))
        vbur = geom2.percent_buried_volume(
            method="MC", scale=1 / 1.1, radius=3
        )
        if not np.isclose(vbur, 100.0 / 27, atol=0.2):
            print("V_bur =", vbur, "expected:", 100.0 / 27)
        self.assertTrue(np.isclose(vbur, 100.0 / 27, atol=0.2))

        geom3 = Geometry(os.path.join(prefix, "ref_files", "vbur2.xyz"))
        vbur = geom2.percent_buried_volume(
            method="MC", scale=1 / 1.1, radius=4
        )
        if not np.isclose(vbur, 100.0 / 64, atol=0.2):
            print("V_bur =", vbur, "expected:", 100.0 / 64)
        self.assertTrue(np.isclose(vbur, 100.0 / 64, atol=0.2))

        # profile.disable()
        # profile.print_stats()

    def test_vbur_lebedev(self):
        """
        tests % volume buried (Lebedev integration)
        """
        # 20, 1454
        geom = Geometry(os.path.join(prefix, "ref_files", "lig_map_3.xyz"))
        vbur = geom.percent_buried_volume(
            method="lebedev", rpoints=20, apoints=1454
        )
        if not np.isclose(vbur, 86.0, atol=0.05):
            print("V_bur =", vbur, "expected:", 86.0)
        self.assertTrue(np.isclose(vbur, 86.0, atol=0.05))

        # 32, 974
        vbur = geom.percent_buried_volume(
            method="lebedev", rpoints=32, apoints=974
        )
        if not np.isclose(vbur, 86.0, atol=0.15):
            print("V_bur =", vbur, "expected:", 86.0)
        self.assertTrue(np.isclose(vbur, 86.0, atol=0.15))

        # 64, 590
        vbur = geom.percent_buried_volume(
            method="lebedev", rpoints=64, apoints=590
        )
        if not np.isclose(vbur, 86.0, atol=0.05):
            print("V_bur =", vbur, "expected:", 86.0)
        self.assertTrue(np.isclose(vbur, 86.0, atol=0.05))

        # 75, 302
        vbur = geom.percent_buried_volume(
            method="lebedev", rpoints=75, apoints=302
        )
        if not np.isclose(vbur, 86.0, atol=0.1):
            print("V_bur =", vbur, "expected:", 86.0)
        self.assertTrue(np.isclose(vbur, 86.0, atol=0.1))

        # 99, 194
        vbur = geom.percent_buried_volume(
            method="lebedev", rpoints=99, apoints=194
        )
        if not np.isclose(vbur, 86.0, atol=0.5):
            print("V_bur =", vbur, "expected:", 86.0)
        self.assertTrue(np.isclose(vbur, 86.0, atol=0.5))

        # 127, 110
        vbur = geom.percent_buried_volume(
            method="lebedev", rpoints=127, apoints=110
        )
        if not np.isclose(vbur, 86.0, atol=0.75):
            print("V_bur =", vbur, "expected:", 86.0)
        self.assertTrue(np.isclose(vbur, 86.0, atol=0.75))

        # 20, 2030
        vbur = geom.percent_buried_volume(
            method="lebedev", rpoints=20, apoints=2030
        )
        if not np.isclose(vbur, 86.0, atol=0.1):
            print("V_bur =", vbur, "expected:", 86.0)
        self.assertTrue(np.isclose(vbur, 86.0, atol=0.1))

        # 20, 2702
        vbur = geom.percent_buried_volume(
            method="lebedev", rpoints=20, apoints=2702
        )
        if not np.isclose(vbur, 86.0, atol=0.15):
            print("V_bur =", vbur, "expected:", 86.0)
        self.assertTrue(np.isclose(vbur, 86.0, atol=0.15))

        # 20, 5810
        vbur = geom.percent_buried_volume(
            method="lebedev", rpoints=20, apoints=5810
        )
        if not np.isclose(vbur, 86.0, atol=0.1):
            print("V_bur =", vbur, "expected:", 86.0)
        self.assertTrue(np.isclose(vbur, 86.0, atol=0.1))

        geom2 = Geometry(os.path.join(prefix, "ref_files", "vbur.xyz"))
        vbur = geom2.percent_buried_volume(
            method="lebedev", rpoints=32, apoints=974, scale=1 / 1.1, radius=2
        )
        if not np.isclose(vbur, 100.0 / 8, atol=0.1):
            print("V_bur =", vbur, "expected:", 100.0 / 8)
        self.assertTrue(np.isclose(vbur, 100.0 / 8, atol=0.1))

    # geometry manipulation
    def test_get_fragment(self):
        mol = Geometry(TestGeometry.benz_NO2_Cl)

        # get Cl using name
        frag = mol.get_fragment("11", "4", copy=False)
        v_frag = mol.atoms[10:11]
        self.assertSequenceEqual(frag, v_frag)

        # get ring without NO2 using atoms
        frag = mol.get_fragment(mol.atoms[0], mol.atoms[11], copy=False)
        v_frag = mol.atoms[:11]
        self.assertSequenceEqual(sorted(frag), sorted(v_frag))

        # get fragment as Geometry()
        frag = mol.get_fragment(mol.atoms[0], mol.atoms[11], as_object=True)
        self.assertIsInstance(frag, Geometry)

    def test_remove_fragment(self):
        benzene = Geometry(TestGeometry.benzene)
        benz_Cl = Geometry(TestGeometry.benz_Cl)
        Et_NO2 = Geometry(TestGeometry.Et_NO2)

        # remove NO2 group using atom name
        mol = Geometry(TestGeometry.benz_NO2_Cl)
        mol.remove_fragment("12", "1")
        self.assertEqual(mol, benz_Cl)
        # remove Cl using elements
        mol.remove_fragment("Cl", "C")
        self.assertEqual(mol, benzene)

        # multiple start atoms
        mol = Geometry(TestGeometry.benz_NO2_Cl)
        mol.remove_fragment(["3", "6"], ["2", "1"])
        self.assertEqual(mol, Et_NO2)

    def test_coord_shift(self):
        benzene = Geometry(TestGeometry.benzene)
        mol = Geometry(TestGeometry.benzene)

        # shift all atoms
        vector = np.array([0, 3.2, -1.0])
        for a in benzene.atoms:
            a.coords += vector
        mol.coord_shift([0, 3.2, -1.0])
        self.assertTrue(np.linalg.norm(benzene.coords - mol.coords) == 0)

        # shift some atoms
        vector = np.array([0, -3.2, 1.0])
        for a in benzene.atoms[0:5]:
            a.coords += vector
        mol.coord_shift([0, -3.2, 1.0], [str(i) for i in range(1, 6)])
        self.assertTrue(np.linalg.norm(benzene.coords - mol.coords) == 0)

    def test_change_distance(self):
        def validate_distance(before, after, moved):
            dist = np.linalg.norm(after.coords - before.coords)
            return abs(dist - moved)

        threshold = 10 ** (-8)
        mol = Geometry(TestGeometry.benz_NO2_Cl)
        a1 = mol.atoms[0]  # C1
        a3 = mol.atoms[5]  # attached to C1
        a2 = mol.atoms[11]  # N12
        a4 = mol.atoms[12]  # attached to N12

        # set distance and move fragments
        a3_before = copy(a3)
        a4_before = copy(a4)
        dist_before = np.linalg.norm(a1.coords - a2.coords)
        mol.change_distance(a1, a2, dist=2.00)
        dist = np.linalg.norm(a1.coords - a2.coords)
        self.assertTrue(is_close(dist, 2.00, threshold))
        self.assertTrue(validate_distance(a3_before, a3, dist - dist_before))
        self.assertTrue(validate_distance(a4_before, a4, dist - dist_before))

        # adjust_distance and move fragments
        a3_before = copy(a3)
        a4_before = copy(a4)
        dist_before = np.linalg.norm(a1.coords - a2.coords)
        mol.change_distance(a1, a2, dist=-0.5, adjust=True)
        self.assertTrue(is_close(a1.dist(a2), 1.5, threshold))
        self.assertTrue(validate_distance(a3_before, a3, -0.5))
        self.assertTrue(validate_distance(a4_before, a4, -0.5))

        # set and fix atom 1
        a1_before = copy(a1)
        mol.change_distance(a1, a2, dist=2.00, fix=1)
        self.assertTrue(is_close(a1.dist(a2), 2.00, threshold))
        self.assertTrue(a1_before.dist(a1) < threshold)

        # adjust and fix atom 2
        a2_before = copy(a2)
        mol.change_distance(a1, a2, dist=-0.5, adjust=True, fix=2)
        self.assertTrue(is_close(a1.dist(a2), 1.5, threshold))
        self.assertTrue(a2_before.dist(a2) < threshold)

        # set and don't move fragments
        a3_before = copy(a3)
        a4_before = copy(a4)
        dist_before = np.linalg.norm(a1.coords - a2.coords)
        mol.change_distance(a1, a2, dist=2.00, as_group=False)
        dist = np.linalg.norm(a1.coords - a2.coords)
        self.assertTrue(is_close(a1.dist(a2), 2.00, threshold))
        self.assertTrue(validate_distance(a3_before, a3, dist - dist_before))
        self.assertTrue(validate_distance(a4_before, a4, dist - dist_before))

    def test_change_angle(self):
        def diff(a, b):
            return abs(a - b)

        mol = Geometry(TestGeometry.benz_NO2_Cl)

        # set angle
        mol.change_angle("13", "12", "14", np.pi / 2, fix=1)
        angle = mol.angle("13", "12", "14")
        self.assertTrue(diff(np.rad2deg(angle), 90) < 10 ** -8)

        # change angle
        mol.change_angle(
            "13", "12", "14", 30, fix=3, adjust=True, radians=False
        )
        angle = mol.angle("13", "12", "14")
        self.assertTrue(diff(np.rad2deg(angle), 120) < 10 ** -8)

        mol.change_angle("13", "12", "14", -30, adjust=True, radians=False)
        angle = mol.angle("13", "12", "14")
        self.assertTrue(diff(np.rad2deg(angle), 90) < 10 ** -8)

    def test_change_dihedral(self):
        mol = Geometry(TestGeometry.benz_NO2_Cl)
        atom_args = ("13", "12", "1", "6")
        original_dihedral = mol.dihedral(*atom_args)

        # adjust dihedral by 30 degrees
        mol.change_dihedral(*atom_args, 30, radians=False, adjust=True)
        self.assertTrue(
            is_close(
                mol.dihedral(*atom_args), original_dihedral + np.deg2rad(30)
            )
        )

        # set dihedral to 60 deg
        mol.change_dihedral(*atom_args, 60, radians=False)
        self.assertTrue(is_close(mol.dihedral(*atom_args), np.deg2rad(60)))

        # adjust using just two atoms
        mol.change_dihedral("12", "1", -30, radians=False, adjust=True)
        self.assertTrue(is_close(mol.dihedral(*atom_args), np.deg2rad(30)))

    def test_substitute(self):
        ref = Geometry(TestGeometry.benz_NO2_Cl)
        mol = Geometry(TestGeometry.benzene)

        mol.substitute(Substituent("NO2"), "12")
        mol.substitute(Substituent("Cl"), "11")

        self.assertTrue(validate(mol, ref))

    def test_close_ring(self):
        mol = Geometry(TestGeometry.benzene)

        ref1 = Geometry(TestGeometry.naphthalene)
        mol1 = mol.copy()
        mol1.ring_substitute(["7", "8"], Ring("benzene"))
        self.assertTrue(validate(mol1, ref1, thresh="loose"))

        ref2 = Geometry(TestGeometry.tetrahydronaphthalene)
        mol2 = mol.copy()
        mol2.ring_substitute(["7", "8"], Ring("cyclohexane"))
        rmsd = mol2.RMSD(ref2, align=True, sort=True)
        self.assertTrue(rmsd < rmsd_tol(ref2, superLoose=True))

        mol3 = Geometry(TestGeometry.naphthalene)
        ref3 = Geometry(TestGeometry.pyrene)
        targets1 = mol3.find(["9", "15"])
        targets2 = mol3.find(["10", "16"])
        mol3.ring_substitute(targets1, Ring("benzene"))
        mol3.ring_substitute(targets2, Ring("benzene"))
        rmsd = mol3.RMSD(ref3, align=True, sort=True)
        self.assertTrue(rmsd < rmsd_tol(ref3, superLoose=True))

    def test_change_element(self):
        mol = Geometry(TestGeometry.benzene)

        ref = Geometry(TestGeometry.pyridine)
        mol.change_element("1", "N", adjust_hydrogens=True)
        self.assertTrue(validate(mol, ref, thresh="loose"))

    def test_map_ligand(self):
        monodentate = Component(TestGeometry.monodentate)
        tridentate = Component(TestGeometry.tridentate)
        debug = False

        # import cProfile
        #
        # profile = cProfile.Profile()
        # profile.enable()

        """
        #TODO: get a reference file for this
        # two monodentate -> bidentate
        ptco4 = TestGeometry.ptco4.copy()
        ptco4.map_ligand('EDA', ["3", "5"])
        """

        # bidentate -> monodentate, none
        ref = Geometry(os.path.join(prefix, "ref_files", "lig_map_1.xyz"))
        tm_simple = Geometry(TestGeometry.tm_simple)
        tm_simple.map_ligand(monodentate.copy(), ["35"])
        self.assertTrue(
            validate(
                tm_simple, ref, heavy_only=True, thresh="loose", debug=debug
            )
        )

        # bidentate -> two monodentate
        ref = Geometry(os.path.join(prefix, "ref_files", "lig_map_2.xyz"))
        tm_simple = Geometry(TestGeometry.tm_simple)
        tm_simple.map_ligand([monodentate.copy(), "ACN"], ["35", "36"])
        self.assertTrue(
            validate(
                tm_simple, ref, thresh="loose", heavy_only=True, debug=debug
            )
        )

        # bidentate -> bidentate
        ref = Geometry(os.path.join(prefix, "ref_files", "lig_map_3.xyz"))
        tm_simple = Geometry(TestGeometry.tm_simple)
        tm_simple.map_ligand("S-tBu-BOX", ["35", "36"])
        self.assertTrue(
            validate(
                tm_simple, ref, thresh="loose", heavy_only=True, debug=debug
            )
        )

        # tridentate -> tridentate
        ref = Geometry(os.path.join(prefix, "ref_files", "lig_map_4.xyz"))
        org_tri = Geometry(TestGeometry.org_tri)
        org_tri.map_ligand(tridentate, ["30", "28", "58"])
        self.assertTrue(
            validate(
                org_tri, ref, thresh="loose", heavy_only=True, debug=debug
            )
        )

        # tridentate -> monodentate + bidentate -> tridentate
        ref = Geometry(os.path.join(prefix, "ref_files", "lig_map_6.xyz"))
        org_tri = Geometry(TestGeometry.org_tri)
        org_tri.map_ligand(["EDA", "ACN"], ["30", "28", "58"])
        self.assertTrue(
            validate(
                org_tri, ref, thresh="loose", heavy_only=True, debug=debug
            )
        )

        ref = Geometry(os.path.join(prefix, "ref_files", "lig_map_7.xyz"))
        org_tri = Geometry(os.path.join(prefix, "ref_files", "lig_map_6.xyz"))
        org_tri.map_ligand(tridentate, ["10", "11", "2"])
        self.assertTrue(
            validate(
                org_tri, ref, thresh="loose", heavy_only=True, debug=debug
            )
        )

        # bidentate -> two bulky monodentate
        ref = Geometry(os.path.join(prefix, "ref_files", "lig_map_5.xyz"))
        tm_simple = Geometry(TestGeometry.tm_simple)
        tm_simple.map_ligand(["iPr-NC3C"] * 2, ["35", "36"])
        self.assertTrue(
            validate(
                tm_simple, ref, thresh="loose", heavy_only=True, debug=debug
            )
        )

        # profile.disable()
        # profile.print_stats()


def suite():
    suite = unittest.TestSuite()
    # suite.addTest(TestGeometry("test_vbur_lebedev"))
    # suite.addTest(TestGeometry("test_examine_constraints"))
    # suite.addTest(TestGeometry("test_detect_components"))
    suite.addTest(TestGeometry("test_fix_comment"))
    # suite.addTest(TestGeometry("test_RMSD"))
    return suite


ONLYSOME = False

if __name__ == "__main__" and ONLYSOME:
    runner = unittest.TextTestRunner()
    runner.run(suite())
elif __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import argparse
import numpy as np

from AaronTools.geometry import Geometry
from AaronTools.fileIO import FileReader

def parse_mode_str(s, t):
    """split mode string into modes and mode combos
    e.g.
    t=int, 1,2+3,4 -> [[0], [1,2], [3]]
    t=float 0.1,0.05+0.03,0.07 -> [[0.1], [0.05, 0.03], [0.07]]"""

    #the way this is being used is if t is int, we are changing 1-indexed things to 0-index
    #if t is float, were going to use the result to scale a normal mode (don"t subtract 1)

    if t is not int and t is not float:
        raise TypeError("can only parse mode string into ints or floats, not %s" % repr(t))

    modes = s.split(",")
    out_modes = []
    for mode in modes:
        out_modes.append([])
        for combo in mode.split("+"):
            if t is int:
                out_modes[-1].append(int(combo)-1)
            elif t is float:
                out_modes[-1].append(float(combo))

    return out_modes

freqbild_parser = argparse.ArgumentParser(
    description="print Chimera bild file with vectors for the specified normal modes to std out",
    formatter_class=argparse.RawTextHelpFormatter
)
freqbild_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    default=None,
    help="a frequency job output file"
)

freqbild_parser.add_argument(
    "-m", "--mode",
    type=str,
    nargs=1,
    default=None,
    required=False,
    metavar="1,2+3,4",
    dest="mode_str",
    help="mode(s) to print (1-indexed)\n" +
    "Default is to print all imaginary modes separately\n" +
    "- comma (,) delimited modes will be printed separately\n" +
    "- plus (+) delimited modes will be combined"
)

freqbild_parser.add_argument(
    "-s", "--scale",
    type=str,
    nargs=1,
    default=None,
    required=False,
    dest="scale",
    metavar="max displacement",
    help="scale the longest vector to be this many Angstroms long\n" +
    "default is 1.5\nmay be delimited in accordance with the --mode option"
)

freqbild_parser.add_argument(
    "-r", "--remove-mass",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="mass_weight",
    help="remove mass-weighting from normal modes"
)

freqbild_parser.add_argument(
    "-c", "--color",
    type=str,
    nargs="+",
    default=["green"],
    required=False,
    dest="color",
    metavar=("BILD 1 color", "BILD 2 color"),
    help="color of vectors"
)

args = freqbild_parser.parse_args()

fr = FileReader(args.infile, just_geom=False)
geom = Geometry(fr)

if args.mode_str is None:
    #if not modes were requested, print all the imaginary ones
    modes = [
        [i] for i, freq in enumerate(fr.other["frequency"].data) if freq.frequency < 0
    ]
else:
    #otherwise, split the modes on delimiters
    modes = parse_mode_str(args.mode_str[0], int)

if args.scale is None:
    scale = [[1.5]*len(mode) for mode in modes]
else:
    scale = parse_mode_str(args.scale[0], float)

color = args.color
colors = len(color)
while colors < len(modes):
    color.extend(args.color)
    colors = len(color)

#output is the string of everything we"ll print
output = ""

for i, mode in enumerate(modes):
    output += ".color %s\n.comment " % args.color[i]
    dX = np.zeros((len(geom.atoms), 3))
    #figure out how much we"ll have to scale each mode
    for j, combo in enumerate(mode):
        output += "%f cm^-1" % fr.other["frequency"].data[combo].frequency
        max_norm = 0
        for k, v in enumerate(fr.other["frequency"].data[combo].vector):
            if args.mass_weight:
                n = np.linalg.norm(v) * geom.atoms[k].mass()
            else:
                n = np.linalg.norm(v)
            if n > max_norm:
                max_norm = n

        # scale this mode by 1.5 (or whatever the user asked for)/max_norm
        x_factor = scale[i][j]/max_norm
        dX += x_factor * fr.other["frequency"].data[combo].vector

        output += " x %.2f " % x_factor

    output += "\n"

    for n in range(0, len(geom.atoms)):
        # scale the vector for each atom and add it to output
        if args.mass_weight:
            dX[n] *= geom.atoms[n].mass()

        v_len = np.linalg.norm(dX[n])

        # we also scale the cone part of the arrow
        start = [x for x in geom.atoms[n].coords]
        stop = [x for x in geom.atoms[n].coords + dX[n]]
        head_len = [v_len / (v_len + 0.75)]

        if v_len > 0.1:
            output += (
                ".arrow %10.6f %10.6f %10.6f   " % tuple(start) +
                "%10.6f %10.6f %10.6f   " % tuple(stop) +
                "0.02 0.05 %5.3f\n" % tuple(head_len)
            )

print(output.rstrip())



"""for handling the change in structure of a series of geometries"""
import numpy as np
from scipy.integrate import quad as integrate


class Pathway:
    """
    interpolating between multiple Geometries

    Attributes:
        geometry - structure for which coordinates are interpolated
        num_geom - number of geometries
        n_cart - number of Cartesian coordinates
        basis -  matrix representation of basis used to interpolate between geometries
        basis_inverse - inverse of basis
        region_length - length of each spline subregion
    """

    def __init__(
            self,
            geometry,
            coordinate_array,
            other_vars=None,
            basis=None,
            mass_weighted=False,
    ):
        """
        geometry - Geometry()
        coordinate_array - np.array(float, shape=(N, n_atoms, 3))
                           coordinates for the geometry at N different points
        other_vars - dict('variable name':[float])
                     dictionary of other variables (e.g. energy)
        basis - list(np.array(float, shape=(n_atoms,3)))
                coordinate displacement matrices (shape n_atoms x 3)
        mass_weighted - bool, True if supplied modes are mass-weighted
        """
        self.geometry = geometry
        self.coordsets = coordinate_array
        if other_vars:
            self.other_vars = other_vars
        else:
            self.other_vars = {}

        if basis is None:
            # if no modes are given, Cartesian coordinates are used
            self.gen_basis([], mass_weighted)
        else:
            self.gen_basis(basis, mass_weighted)

        # set up coordinate and energy functions
        self.gen_func()

    @property
    def num_geom(self):
        """returns number of input coordinate sets"""
        return self.coordsets.shape[0]

    def gen_basis(self, modes, remove_mass_weight=False):
        """
        modes - list of coordinate displacement matrices
        remove_mass_weight  bool, True to remove mass-weighting from displacement coordinates
        """

        basis = []
        n_cart = 3 * len(self.geometry.atoms)

        # go through each of the supplied coordinate displacement matrices
        # remove mass-weighting if needed, and reshape
        for mode in modes:
            no_mass_mode = []
            for i, atom in enumerate(self.geometry.atoms):
                if remove_mass_weight:
                    no_mass_mode.append(mode[i] * atom.mass())
                else:
                    no_mass_mode.append(mode[i])

            basis.append(np.array(np.reshape(no_mass_mode, n_cart)))

        # we need n_cart basis vectors
        if len(basis) < n_cart:
            if basis:
                raise RuntimeError(
                    "number of basis vectors (%i) is less than 3N (%i)"
                    % (len(basis), n_cart)
                )

            # if we don't have any, this is equivalent to using each atom's
            # x, y, and z coordinates as our basis
            basis = np.identity(n_cart)

        basis = np.transpose(basis)

        self.basis = basis
        self.basis_inverse = np.linalg.inv(basis)

    def gen_func(self):
        """
        generate Cartesian-displacement-representation coordinate and
        miscellaneous variable interpolation functions
        sets self.coord_func, self.dcoord_func_dt, self.E_func, and self.dE_func_dt
        """

        basis_rep = []
        n_cart = 3 * len(self.geometry.atoms)
        for xyz in self.coordsets:
            d_xyz = xyz - self.coordsets[0]
            a = Pathway.dxyz_to_q(d_xyz, self.basis_inverse)
            basis_rep.append(np.reshape(a, (n_cart, 1)))

        basis_rep = np.reshape(np.array(basis_rep), (n_cart * self.num_geom, 1))

        # get cubic spline coefficients for the subregions
        # solved by mat * coord = basis -> coord = mat^-1 * basis
        mat = Pathway.get_splines_mat(self.num_geom)
        # basis is a matrix with basis rep. coefficients, or zeros for derivative rows in M
        basis = np.zeros((4 * (self.num_geom - 1), n_cart))

        for i in range(0, self.num_geom - 1):
            for j in range(0, n_cart):
                basis[2 * i][j] = basis_rep[i * n_cart + j][0]
                basis[2 * i + 1][j] = basis_rep[(i + 1) * n_cart + j][0]

        mat_i = np.linalg.inv(mat)
        coord = np.dot(mat_i, basis)

        # get arc length for each region
        arc_length = Pathway.get_arc_length(coord)

        # region_length = [simpson(arc_length, m, m+1) for m in range(0, self.num_geom-1)]
        region_length = [
            integrate(arc_length, m, m + 1)[0] for m in range(0, self.num_geom - 1)
        ]
        self.region_length = region_length

        # set self's coordinate function
        # coordinates are coefficients for Cartesian displacement representation
        self.coord_func, self.dcoord_func_dt = self.get_coord_func(
            coord, region_length
        )
        self.var_func = {}
        self.dvar_func_dt = {}
        for var in self.other_vars:
            c_var = np.dot(mat_i, Pathway.get_splines_vector(self.other_vars[var]))
            self.var_func[var], self.dvar_func_dt[var] = Pathway.get_var_func(
                c_var, region_length
            )

    def geom_func(self, t):
        """returns a Geometry from the interpolated pathway at point t
        t       float       point on pathway {t|0 <= t <= 1}"""
        geom = self.geometry.copy()
        geom.update_geometry(self.coords_func(t))
        return geom

    def coords_func(self, t):
        """returns Cartesian coordinates for the geometry at point t"""
        Q = self.coord_func(t)
        return self.q_to_xyz(Q)

    def dcoords_dt_func(self, t):
        """returns derivative of Cartesian coordinates for the geometry at point t"""
        Q = self.dcoord_func_dt(t)
        return self.q_to_xyz(Q)

    def get_coord_func(self, coord, region_length):
        """
        returns function for Cartesian displacement coordinate as a function of t (t [0, 1])
        and a derivative of this function
        coord - array-like(float, shape = (4*n_subregions, n_cart))
                matrix of cubic polynomial coefficients
        region_length - array-like(float)
                        arc length of each subregion
        """
        n_cart = 3 * len(self.geometry.atoms)

        def coord_fun(t):
            # map input t to s and region number
            s, r = Pathway.t_to_s(t, region_length)

            # evaluate polynomial
            q = np.array(
                [
                    coord[4 * r][i] * (s - r) ** 3
                    + coord[4 * r + 1][i] * (s - r) ** 2
                    + coord[4 * r + 2][i] * (s - r)
                    + coord[4 * r + 3][i]
                    for i in range(0, n_cart)
                ]
            )

            return q

        def dcoord_dt(t):
            s, r = Pathway.t_to_s(t, region_length)

            q = np.array(
                [
                    3 * coord[4 * r][i] * (s - r) ** 2
                    + 2 * coord[4 * r + 1][i] * (s - r)
                    + coord[4 * r + 2][i]
                    for i in range(0, n_cart)
                ]
            )

            return q

        return coord_fun, dcoord_dt

    @staticmethod
    def get_var_func(c_var, region_length):
        """just like get_coord_func, but for other variables"""

        def var_func(t):
            s, r = Pathway.t_to_s(t, region_length)

            var = (
                c_var[4 * r] * (s - r) ** 3
                + c_var[4 * r + 1] * (s - r) ** 2
                + c_var[4 * r + 2] * (s - r)
                + c_var[4 * r + 3]
            )

            return var

        def dvardt_func(t):
            s, r = Pathway.t_to_s(t, region_length)

            dvar = (
                3 * c_var[4 * r] * (s - r) ** 2
                + 2 * c_var[4 * r + 1] * (s - r)
                + c_var[4 * r + 2]
            )

            return dvar

        return var_func, dvardt_func

    @staticmethod
    def dxyz_to_q(dxyz, basis_inverse):
        """
        converts Cartesian changes (dxyz) to whatever basis set basis_inverse
        is (normal mode displacements/Cartesian)
        returns a vector containing the coefficients of each basis matrix
        """
        q = np.reshape(dxyz, 3 * len(dxyz))
        a = np.dot(basis_inverse, q)
        return a

    def q_to_xyz(self, current_q):
        """converts coordinates for self.basis to Cartesian"""
        coords = self.coordsets[0].copy()
        for i, mode in enumerate(np.transpose(self.basis)):
            coords += current_q[i] * np.reshape(
                mode, (len(self.geometry.atoms), 3)
            )

        return coords

    @staticmethod
    def get_splines_mat(n_nodes):
        """generate matrix for fitting cubic splines to data
        matrix is 4*n_regions x 4*n_regions (n_regions = n_nodes-1)
        additional contraints (that might not be valid) as that
        the first derivatives are 0 at both ends of the interpolation
        region (e.g. f'(0) = 0 and f'(1) = 0)"""
        mat = np.zeros((4 * (n_nodes - 1), 4 * (n_nodes - 1)))

        # function values are equal where regions meet
        for i in range(0, n_nodes - 1):
            mat[2 * i][4 * (i + 1) - 1] = 1
            for k in range(4 * i, 4 * (i + 1)):
                mat[2 * i + 1][k] = 1

        # 1st derivatives are equal where regions meet
        for i in range(0, n_nodes - 2):
            j = 2 * (n_nodes - 1) + i
            mat[j][4 * i] = 3
            mat[j][4 * i + 1] = 2
            mat[j][4 * i + 2] = 1
            mat[j][4 * i + 6] = -1

        # 2nd derivatives are equal where regions meet
        for i in range(0, n_nodes - 2):
            j = 3 * (n_nodes - 1) - 1 + i
            mat[j][4 * i] = 6
            mat[j][4 * i + 1] = 2
            mat[j][4 * i + 5] = -2

        # 1st derivatives are 0 at the ends
        mat[-2][2] = 1
        mat[-1][-2] = 1
        mat[-1][-3] = 2
        mat[-1][-4] = 3

        return mat

    @staticmethod
    def get_splines_vector(data):
        """organize data into a vector that can be used with cubic splines matrix"""
        n_regions = len(data) - 1

        v = np.zeros(4 * (n_regions))

        for i in range(0, n_regions):
            v[2 * i] = data[i]
            v[2 * i + 1] = data[i + 1]

        return v

    @staticmethod
    def t_to_s(t, region_length):
        """
        maps t ([0, 1]) to s (changes linearly with displacement coordinates
        need to map b/c cubic splines polynomials generated for interpolation subregions
        should be given an input between 0 and 1, no matter where they are on the
        whole interpolation
        s               float           point on interpolation arc
        region_length   list(float)     arc length of each region

        returns s, r
        r is the region number
        s is the point in that region
        """
        n_regions = len(region_length)
        path_length = sum(region_length)
        region_start = [
            sum(region_length[:i]) for i in range(0, len(region_length))
        ]
        u = t * path_length
        r = 0
        for l in range(0, n_regions):
            if u > region_start[l]:
                r = l

        s = r + (u - region_start[r]) / region_length[r]
        return s, r

    @staticmethod
    def s_to_t(s, region_length):
        """
        map s (changes linearly with displacement coordinate) to t (ranges from 0 to 1)
        s               float           point on interpolation arc
        region_length   list(float)     arc length of each region

        returns t   float
        """
        n_regions = len(region_length)
        path_length = sum(region_length)
        region_start = [
            sum(region_length[:i]) for i in range(0, len(region_length))
        ]
        r = int(s)
        while r >= (n_regions):
            r -= 1
        u = (s - r) * region_length[r] + region_start[r]
        t = u / path_length
        return t

    @staticmethod
    def get_arc_length(coord):
        """
        returns a function that can be integrated to determine the arc length
        of interpolation splines before normalization
        coord - array-like(float, shape = (4*n_subregions, n_cart))
                matrix of cubic polynomial coefficients

        returns function(s)
        """
        n_cart = len(coord[0])

        def unnormalized_func(s):
            r = int(s)
            if r == s:
                r = int(s - 1)

            f = 0
            for i in range(0, n_cart):
                f += (
                    3 * coord[4 * r][i] * (s - r) ** 2
                    + 2 * coord[4 * r + 1][i] * (s - r)
                    + coord[4 * r + 2][i]
                ) ** 2

            return np.sqrt(f)

        return unnormalized_func



import sys
import argparse

import numpy as np

from AaronTools.fileIO import FileReader, read_types
from AaronTools.geometry import Geometry
from AaronTools.symmetry import (
    PointGroup,
    InversionCenter, 
    ProperRotation,
    ImproperRotation,
    MirrorPlane,
)
from AaronTools.utils.utils import perp_vector, glob_files

pg_parser = argparse.ArgumentParser(
    description="print point group",
    formatter_class=argparse.RawTextHelpFormatter
)

pg_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file"
)

pg_parser.add_argument(
    "-o", "--output",
    type=str,
    default=False,
    required=False,
    dest="outfile",
    help="output destination \nDefault: stdout"
)

pg_parser.add_argument(
    "-if", "--input-format",
    type=str,
    default=None,
    dest="input_format",
    choices=read_types,
    help="file format of input - xyz is assumed if input is stdin"
)

pg_parser.add_argument(
    "-t", "--tolerance",
    default=0.1,
    type=float,
    dest="tolerance",
    help="tolerance for determining if a symmetry element is valid\n"
    "for the input structure(s)\nDefault: 0.1"
)

pg_parser.add_argument(
    "-a", "--axis-tolerance",
    default=0.01,
    type=float,
    dest="rotation_tolerance",
    help="tolerance for determining if two axes are coincident or orthogonal"
    "\nDefault: 0.01"
)

pg_parser.add_argument(
    "-n", "--max-n",
    default=6,
    type=int,
    dest="max_n",
    help="max. order for proper rotation axes (improper rotations can be 2x this)"
    "\nDefault: 6"
)

pg_parser.add_argument(
    "-v", "--verbose",
    action="store_true",
    default=False,
    dest="print_eles",
    help="print all symmetry elements",
)

pg_parser.add_argument(
    "-b", "--bild",
    action="store_true",
    default=False,
    dest="print_bild",
    help="print Chimera(X) bild file to display various symmetry elements",
)

args = pg_parser.parse_args()

s = ""
for f in glob_files(args.infile, parser=pg_parser):
    if isinstance(f, str):
        if args.input_format is not None:
            infile = FileReader((f, args.input_format, None), just_geom=True)
        else:
            infile = FileReader(f, just_geom=True)
    else:
        if args.input_format is not None:
            infile = FileReader(("from stdin", args.input_format[0], f), just_geom=True)
        else:
            if len(sys.argv) >= 1:
                infile = FileReader(("from stdin", "xyz", f), just_geom=True)

    geom = Geometry(infile)
    pg = PointGroup(
        geom,
        tolerance=args.tolerance,
        rotation_tolerance=args.rotation_tolerance,
        max_rotation=args.max_n,
    )
    if args.print_bild:
        inv = ""
        mirror = ""
        prots = ""
        irots = ""
        for ele in sorted(pg.elements, reverse=True):
            if isinstance(ele, InversionCenter):
                inv += ".note %s\n" % repr(ele)
                inv += ".color plum\n"
                inv += ".sphere   %.5f  %.5f  %.5f  0.3\n" % tuple(pg.center)

            elif isinstance(ele, ProperRotation):
                prots += ".note %s\n" % repr(ele)
                prots += ".color red\n"
                prots += ".arrow   %.5f  %.5f  %.5f  " % tuple(pg.center)
                end = pg.center + ele.n * np.sqrt(ele.exp) * ele.axis
                prots += "%.5f  %.5f  %.5f  0.05\n"  % tuple(end)

            elif isinstance(ele, ImproperRotation):
                irots += ".note %s\n" % repr(ele)
                irots += ".color blue\n"
                irots += ".arrow   %.5f  %.5f  %.5f  " % tuple(pg.center)
                end = pg.center + np.sqrt(ele.n) * np.sqrt(ele.exp) * ele.axis
                irots += "%.5f  %.5f  %.5f  0.05\n"  % tuple(end)
                irots += ".transparency 25\n"
                z = ele.axis
                x = perp_vector(z)
                y = np.cross(x, z)
                for angle in np.linspace(0, 2 * np.pi, num=25):
                    pt2 = ele.n ** 0.9 * x * np.cos(angle)
                    pt2 += ele.n ** 0.9 * y * np.sin(angle)
                    pt2 += pg.center
                    if angle > 0:
                        irots += ".polygon  %6.3f  %6.3f  %6.3f" % tuple(pt1)
                        irots += "     %6.3f  %6.3f  %6.3f" % tuple(pg.center)
                        irots += "     %6.3f  %6.3f  %6.3f" % tuple(pt2)
                        irots += "\n"
                    pt1 = pt2

            elif isinstance(ele, MirrorPlane):
                mirror += ".note %s\n" % repr(ele)
                if ele.label is None:
                    mirror += ".color purple\n"
                elif ele.label == "h":
                    mirror += ".color black\n"
                elif ele.label == "v":
                    mirror += ".color chocolate\n"
                elif ele.label == "d":
                    mirror += ".color teal\n"
                mirror += ".transparency 40\n"
                z = ele.axis
                x = perp_vector(z)
                y = np.cross(x, z)
                for angle in np.linspace(0, 2 * np.pi, num=25):
                    pt2 = 5 * x * np.cos(angle)
                    pt2 += 5 * y * np.sin(angle)
                    pt2 += pg.center
                    if angle > 0:
                        mirror += ".polygon  %6.3f  %6.3f  %6.3f" % tuple(pt1)
                        mirror += "     %6.3f  %6.3f  %6.3f" % tuple(pg.center)
                        mirror += "     %6.3f  %6.3f  %6.3f" % tuple(pt2)
                        mirror += "\n"
                    pt1 = pt2

        if args.outfile:
            with open(args.outfile, "w") as f:
                f.write("\n".join([inv, prots, irots, mirror]))
        else:
            if inv:
                print(inv)
            if prots:
                print(prots)
            if irots:
                print(irots)
            if mirror:
                print(mirror)

    else:
        s += "%s: %s\n" % (f, pg.name)
        if args.print_eles:
            for ele in sorted(pg.elements, reverse=True):
                s += "\t%s\n" % repr(ele)

if not args.print_bild:
    if args.outfile:
        with open(args.outfile, "w") as f:
            f.write(s)
    else:
        print(s)



#!/usr/bin/env python3

import sys
import argparse
import copy

import numpy as np
import matplotlib.pyplot as plt

from AaronTools.geometry import Geometry
from AaronTools.fileIO import FileReader, read_types
from AaronTools.utils.utils import rotation_matrix, get_filename, glob_files


steric_parser = argparse.ArgumentParser(
    description="create a steric map for a ligand",
    formatter_class=argparse.RawTextHelpFormatter
)

steric_parser.add_argument(
    "infile",
    metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file",
)

steric_parser.add_argument(
    "-o",
    "--output",
    type=str,
    default=False,
    required=False,
    dest="outfile",
    help="output destination\n" +
    "$INFILE will be replaced with the name of the input file\n" +
    "Default: show plot",
)

steric_parser.add_argument(
    "-if",
    "--input-format",
    type=str,
    default=None,
    dest="input_format",
    choices=read_types,
    help="file format of input - xyz is assumed if input is stdin"
)

steric_parser.add_argument(
    "-k",
    "--key-atoms",
    default=None,
    required=False,
    dest="key",
    help="atoms coordinated to the center\n" +
    "these atoms\" ligands will be shown on the map",
)

steric_parser.add_argument(
    "-c",
    "--center",
    action="append",
    default=None,
    required=False,
    dest="center",
    help="atom the sphere is centered on\n" +
    "Default: detect metal center (centroid of all metals if multiple are present)",
)

steric_parser.add_argument(
    "-v",
    "--vdw-radii",
    default="umn",
    choices=["umn", "bondi"],
    dest="radii",
    help="VDW radii to use in calculation\n" + 
    "umn: main group vdw radii from J. Phys. Chem. A 2009, 113, 19, 5806–5812\n" +
    "    (DOI: 10.1021/jp8111556)\n" + 
    "    transition metals are crystal radii from Batsanov, S.S. Van der Waals\n" +
    "    Radii of Elements. Inorganic Materials 37, 871–885 (2001).\n" +
    "    (DOI: 10.1023/A:1011625728803)\n" + 
    "bondi: radii from J. Phys. Chem. 1964, 68, 3, 441–451 (DOI: 10.1021/j100785a001)\n" +
    "Default: umn",
)

steric_parser.add_argument(
    "-r",
    "--radius",
    default=3.5,
    type=float,
    dest="radius",
    help="radius around center\nDefault: 3.5 Ångström"
)

steric_parser.add_argument(
    "-oop",
    "--out-of-plane",
    default=None,
    nargs=3,
    type=float,
    metavar=("x_oop", "y_oop", "z_oop"),
    dest="oop_vector",
    help="list of three numbers defining a vector perpendicular to\n" +
    "the desired steric map",
)

steric_parser.add_argument(
    "-ip",
    "--in-plane",
    default=None,
    nargs=3,
    type=float,
    metavar=("x_ip", "y_ip", "z_ip"),
    dest="ip_vector",
    help="list of three numbers defining the \"y axis\" of the steric map",
)

steric_parser.add_argument(
    "-n", "--number-of-points",
    default=100,
    type=int,
    dest="num_pts",
    help="number of points along x and y axes\nDefault: 100",
)

steric_parser.add_argument(
    "-amin", "--altitude-minimum",
    default=None,
    type=float,
    dest="min",
    help="manually set the lower cutoff of the altitude map",
)

steric_parser.add_argument(
    "-amax",
    "--altitude-maximum",
    default=None,
    type=float,
    dest="max",
    help="manually set the upper cutoff of the altitude map",
)

steric_parser.add_argument(
    "-p", "--projection-shape",
    choices=("circle", "square"),
    default="circle",
    dest="shape",
    help="shape of steric map\n" +
    "note that buried volume values with the square shape are still based\n" +
    "on a sphere around the center\n" +
    "Default: circle",
)

vbur_options = steric_parser.add_argument_group("Buried volume options")
vbur_options.add_argument(
    "-vbur", "--buried-volume",
    nargs="?",
    default=False,
    choices=("Lebedev", "MC"),
    # this allows these choices to be case-insensitive, but we can still
    # show normal upper- and lowercase in the help page
    type=lambda x: x.capitalize() if x.lower() == "lebedev" else x.upper(),
    dest="vbur",
    help="show buried volume in each quadrant using the specified method\n" +
    "Default: do not include %%Vbur",
)

vbur_options.add_argument(
    "-rp", "--radial-points",
    type=int,
    default=20,
    choices=[20, 32, 64, 75, 99, 127],
    dest="rpoints",
    help="number of radial shells for Lebedev integration\n" +
    "lower values are faster, but at the cost of accuracy\n" +
    "Default: 20"
)

vbur_options.add_argument(
    "-ap", "--angular-points",
    type=int,
    default=1454,
    choices=[110, 194, 302, 590, 974, 1454, 2030, 2702, 5810],
    dest="apoints",
    help="number of angular points for Lebedev integration\n" +
    "lower values are faster, but at the cost of accuracy\n" +
    "Default: 1454"
)

vbur_options.add_argument(
    "-i",
    "--minimum-iterations",
    type=int,
    default=25,
    metavar="ITERATIONS",
    dest="min_iter",
    help="minimum iterations - each is a batch of 3000 points\n" +
    "MC will continue after this until convergence criteria are met\n" +
    "Default: 25",
)

vbur_options.add_argument(
    "-s",
    "--scale",
    type=float,
    dest="scale",
    default=1.17,
    help="scale VDW radii by this amount\nDefault: 1.17"
)


args = steric_parser.parse_args()

oop_vector = args.oop_vector
if args.oop_vector is not None:
    oop_vector = np.array(args.oop_vector)

ip_vector = args.ip_vector
if args.ip_vector is not None:
    ip_vector = np.array(args.ip_vector)

if args.vbur is None:
    args.vbur = "Lebedev"

for f in glob_files(args.infile, parser=steric_parser):
    if isinstance(f, str):
        if args.input_format is not None:
            infile = FileReader((f, args.input_format[0], None))
        else:
            infile = FileReader(f, just_geom=False)
    else:
        if args.input_format is not None:
            infile = FileReader(("from stdin", args.input_format[0], f))
        else:
            if len(sys.argv) >= 1:
                infile = FileReader(("from stdin", "xyz", f))

    geom = Geometry(infile)

    x, y, z, min_alt, max_alt, basis, targets = geom.steric_map(
        center=args.center,
        key_atoms=args.key,
        radii=args.radii,
        return_basis=True,
        num_pts=args.num_pts,
        oop_vector=oop_vector,
        ip_vector=ip_vector,
        shape=args.shape,
    )

    if args.ip_vector is None or args.oop_vector is None:
        print(f)

    if args.oop_vector is None:
        z_vec = np.squeeze(basis[:, 2])
        print("out-of-plane vector: %s" % " ".join(["%6.3f" % yi for yi in z_vec]))

    if args.ip_vector is None:
        z_vec = np.squeeze(basis[:, 2])
        y_vec = np.squeeze(basis[:, 1])
        r15 = rotation_matrix(np.deg2rad(15), z_vec)
        yr = y_vec
        for i in range(1, 24):
            yr = np.dot(r15, yr)
            print("in-plane vector rotated by %5.1f degrees: %s" % (
                (15 * i), " ".join(["%6.3f" % yi for yi in yr])
            ))

    if args.min is not None:
        min_alt = args.min

    if args.max is not None:
        max_alt = args.max

    fig, ax = plt.subplots()
    steric_map = ax.contourf(
        x,
        y,
        z,
        extend="min",
        cmap=copy.copy(plt.cm.get_cmap("jet")),
        levels=np.linspace(min_alt, max_alt, num=20),
    )
    steric_map.cmap.set_under("w")
    steric_lines = ax.contour(
        x,
        y,
        z,
        extend="min",
        colors="k",
        levels=np.linspace(min_alt, max_alt, num=20),
    )
    bar = fig.colorbar(steric_map, format="%.1f")
    bar.set_label("altitude (Å)")
    ax.set_aspect("equal")

    if args.vbur:
        vbur = geom.percent_buried_volume(
            center=args.center,
            targets=targets,
            radius=args.radius,
            radii=args.radii,
            scale=args.scale,
            method=args.vbur,
            rpoints=args.rpoints,
            apoints=args.apoints,
            basis=basis,
            min_iter=args.min_iter,
        )

        ax.hlines(0, -args.radius, args.radius, color="k")
        ax.vlines(0, -args.radius, args.radius, color="k")

        vbur_1 = vbur[0] + vbur[7]
        vbur_2 = vbur[1] + vbur[6]
        vbur_3 = vbur[2] + vbur[5]
        vbur_4 = vbur[3] + vbur[4]
        ax.text(+0.7 * args.radius, +0.9 * args.radius, "%.1f%%" % vbur_1)
        ax.text(-0.9 * args.radius, +0.9 * args.radius, "%.1f%%" % vbur_2)
        ax.text(-0.9 * args.radius, -0.9 * args.radius, "%.1f%%" % vbur_3)
        ax.text(+0.7 * args.radius, -0.9 * args.radius, "%.1f%%" % vbur_4)

    if not args.outfile:
        plt.show()
    else:
        outfile = args.outfile
        if "$INFILE" in outfile:
            outfile = outfile.replace("$INFILE", get_filename(f))
        plt.savefig(outfile, dpi=500)



#!/usr/bin/env python3

import sys
import argparse
from warnings import warn

from AaronTools.geometry import Geometry
from AaronTools.fileIO import FileReader, read_types
from AaronTools.finders import NotAny
from AaronTools.utils.utils import glob_files


vbur_parser = argparse.ArgumentParser(
    description="calculated % volume buried in a sphere around a center atom - see Organometallics 2008, 27, 12, 2679–2681",
    formatter_class=argparse.RawTextHelpFormatter
)

vbur_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file"
)

vbur_parser.add_argument(
    "-o", "--output",
    type=str,
    default=False,
    required=False,
    dest="outfile",
    help="output destination \nDefault: stdout"
)

vbur_parser.add_argument(
    "-if", "--input-format",
    type=str,
    default=None,
    dest="input_format",
    choices=read_types,
    help="file format of input - xyz is assumed if input is stdin"
)

vbur_parser.add_argument(
    "-t", "--targets",
    default=None,
    required=False,
    dest="targets",
    help="atoms to consider in calculation\nDefault: use all atoms except the center",
)

vbur_parser.add_argument(
    "-e", "--exclude-atoms",
    default=None,
    required=False,
    dest="exclude_atoms",
    help="atoms to exclude from the calculation\nDefault: exclude no ligand atoms",
)

vbur_parser.add_argument(
    "-c", "--center",
    action="append",
    default=None,
    required=False,
    dest="center",
    help="atom the sphere is centered on\n" +
    "Default: detect metal center (centroid of all metals if multiple are present)",
)

vbur_parser.add_argument(
    "-v", "--vdw-radii",
    default="umn",
    choices=["umn", "bondi"],
    dest="radii",
    help="VDW radii to use in calculation\n" + 
    "umn: main group vdw radii from J. Phys. Chem. A 2009, 113, 19, 5806–5812\n" +
    "    (DOI: 10.1021/jp8111556)\n" + 
    "    transition metals are crystal radii from Batsanov, S.S. Van der Waals\n" +
    "    Radii of Elements. Inorganic Materials 37, 871–885 (2001).\n" +
    "    (DOI: 10.1023/A:1011625728803)\n" + 
    "bondi: radii from J. Phys. Chem. 1964, 68, 3, 441–451 (DOI: 10.1021/j100785a001)\n" +
    "Default: umn",
)

vbur_parser.add_argument(
    "-s", "--scale",
    default=1.17,
    type=float,
    dest="scale",
    help="scale VDW radii by this amount\nDefault: 1.17",
)

vbur_parser.add_argument(
    "-r", "--radius",
    default=3.5,
    type=float,
    dest="radius",
    help="radius around center\nDefault: 3.5 Ångström"
)

vbur_parser.add_argument(
    "-dr", "--scan",
    default=[0, 1],
    nargs=2,
    type=float,
    dest="scan",
    metavar=["dR", "NUMBER"],
    help="calculate %%Vbur with NUMBER different radii, starting with\n"
    "the radius specified with -r/--radius and increasing\n"
    "in increments in dR"
)

vbur_parser.add_argument(
    "-m", "--method",
    default="Lebedev",
    type=lambda x: x.capitalize() if x.lower() == "lebedev" else x.upper(),
    choices=["MC", "Lebedev"],
    dest="method",
    help="integration method - Monte-Carlo (MC) or Lebedev quadrature (Lebedev)\nDefault: Lebedev"
)

grid_options = vbur_parser.add_argument_group("Lebedev integration options")
grid_options.add_argument(
    "-rp", "--radial-points",
    type=int,
    default=20,
    choices=[20, 32, 64, 75, 99, 127],
    dest="rpoints",
    help="number of radial shells for Gauss-Legendre integration\n" +
    "of the radial component\n" +
    "lower values are faster, but at the cost of accuracy\nDefault: 20"
)

grid_options.add_argument(
    "-ap", "--angular-points",
    type=int,
    default=1454,
    choices=[110, 194, 302, 590, 974, 1454, 2030, 2702, 5810],
    dest="apoints",
    help="number of angular points for Lebedev integration\n" +
    "lower values are faster, but at the cost of accuracy\nDefault: 1454"
)

mc_options = vbur_parser.add_argument_group("Monte-Carlo integration options")
mc_options.add_argument(
    "-i", "--minimum-iterations",
    type=int,
    default=25,
    metavar="ITERATIONS",
    dest="min_iter",
    help="minimum iterations - each is a batch of 3000 points\n" +
    "MC will continue after this until convergence criteria are met\n" +
    "Default: 25",
)

args = vbur_parser.parse_args()

s = ""

for f in glob_files(args.infile, parser=vbur_parser):
    if isinstance(f, str):
        if args.input_format is not None:
            infile = FileReader((f, args.input_format[0], None))
        else:
            infile = FileReader(f, just_geom=False)
    else:
        if args.input_format is not None:
            infile = FileReader(("from stdin", args.input_format[0], f))
        else:
            if len(sys.argv) >= 1:
                infile = FileReader(("from stdin", "xyz", f))

    geom = Geometry(infile)

    targets = None
    if args.exclude_atoms and not args.targets:
        targets = (NotAny(args.exclude_atoms))
    elif args.exclude_atoms and args.targets:
        targets = (NotAny(args.exclude_atoms), args.targets)
    elif not args.exclude_atoms and args.targets:
        targets = geom.find(args.targets)
    else:
        targets = NotAny(args.center)

    radius = args.radius
    for i in range(0, int(args.scan[1])):
        try:
            vbur = geom.percent_buried_volume(
                targets=targets,
                center=args.center,
                radius=radius,
                radii=args.radii,
                scale=args.scale,
                method=args.method,
                rpoints=args.rpoints,
                apoints=args.apoints,
                min_iter=args.min_iter,
            )
    
            if i == 0:
                if len(args.infile) > 1:
                    s += "%-20s\n" % (f + ":")
        
                s += "radius\t%Vbur\n"
            s += "%.2f\t%4.1f\n" % (radius, vbur)
        except Exception as e:
            raise RuntimeError("calulation failed for %s: %s" % (f, e))
            break
        
        radius += args.scan[0]
    else:
        s += "\n"

if not args.outfile:
    print(s.strip())
else:
    with open(args.outfile, "a") as f:
        f.write(s.strip())



#!/usr/bin/env python3

import argparse
import sys

from AaronTools.fileIO import FileReader
from AaronTools.utils.utils import get_filename, glob_files

from matplotlib import rcParams
import matplotlib.pyplot as plt

import numpy as np

rcParams["savefig.dpi"] = 300


peak_types = ["pseudo-voigt", "gaussian", "lorentzian", "delta"]
plot_types = ["transmittance", "absorbance", "vcd", "raman"]


def peak_type(x):
    out = [y for y in peak_types if y.startswith(x)]
    if out:
        return out[0]
    raise TypeError(
        "peak type must be one of: %s" % ", ".join(
            peak_types
        )
    )

def plot_type(x):
    out = [y for y in plot_types if y.startswith(x)]
    if out:
        return out[0]
    raise TypeError(
        "plot type must be one of: %s" % ", ".join(
            plot_types
        )
    )


ir_parser = argparse.ArgumentParser(
    description="plot IR spectrum",
    formatter_class=argparse.RawTextHelpFormatter
)

ir_parser.add_argument(
    "infiles", metavar="files",
    type=str,
    nargs="+",
    help="frequency job output file(s)"
)

ir_parser.add_argument(
    "-o", "--output",
    type=str,
    default=None,
    dest="outfile",
    help="output destination\nDefault: show plot",
)

ir_parser.add_argument(
    "-t", "--plot-type",
    type=plot_type,
    choices=plot_types,
    default="transmittance",
    dest="plot_type",
    help="type of plot\nDefault: transmittance",
)


# TODO: figure out more anharmonic options
# anharmonic_options = ir_parser.add_argument_group("anharmonic options")
ir_parser.add_argument(
    "-na", "--harmonic",
    action="store_false",
    default=True,
    dest="anharmonic",
    help="force to use harmonic frequencies when anharmonic data is in the file",
)


peak_options = ir_parser.add_argument_group("peak options")
peak_options.add_argument(
    "-p", "--peak-type",
    type=peak_type,
    choices=peak_types,
    default="pseudo-voigt",
    dest="peak_type",
    help="function for peaks\nDefault: pseudo-voigt",
)

peak_options.add_argument(
    "-m", "--voigt-mixing",
    type=float,
    default=0.5,
    dest="voigt_mixing",
    help="fraction of pseudo-Voigt that is Gaussian\nDefault: 0.5",
)

peak_options.add_argument(
    "-fwhm", "--full-width-half-max",
    type=float,
    default=15.0,
    dest="fwhm",
    help="full width at half max. of peaks\nDefault: 15 cm^-1",
)

ir_parser.add_argument(
    "-s", "--point-spacing",
    default=None,
    type=float,
    dest="point_spacing",
    help="spacing between each x value\n"
    "Default: a non-uniform spacing that is more dense near peaks",
)

scale_options = ir_parser.add_argument_group("scale frequencies")
scale_options.add_argument(
    "-l", "--linear-scale",
    type=float,
    default=0.0,
    dest="linear_scale",
    help="subtract linear_scale * frequency from each mode (i.e. this is 1 - λ)\n"
    "Default: 0 (no scaling)",
)

scale_options.add_argument(
    "-q", "--quadratic-scale",
    type=float,
    default=0.0,
    dest="quadratic_scale",
    help="subtract quadratic_scale * frequency^2 from each mode\n"
    "Default: 0 (no scaling)",
)

ir_parser.add_argument(
    "-nr", "--no-reverse",
    action="store_false",
    default=True,
    dest="reverse_x",
    help="do not reverse x-axis",
)

center_centric = ir_parser.add_argument_group("x-centered interruptions")
center_centric.add_argument(
    "-sc", "--section-centers",
    type=lambda x: [float(v) for v in x.split(",")],
    dest="centers",
    default=None,
    help="split plot into sections with a section centered on each of the specified values\n"
    "values should be separated by commas"
)

center_centric.add_argument(
    "-sw", "--section-widths",
    type=lambda x: [float(v) for v in x.split(",")],
    dest="widths",
    default=None,
    help="width of each section specified by -c/--centers\n"
    "should be separated by commas, with one for each section"
)

minmax_centric = ir_parser.add_argument_group("x-range interruptions")
minmax_centric.add_argument(
    "-r", "--ranges",
    type=lambda x: [[float(v) for v in r.split("-")] for r in x.split(",")],
    dest="ranges",
    default=None,
    help="split plot into sections (e.g. 0-1900,2900-3300)"
)

ir_parser.add_argument(
    "-fw", "--figure-width",
    type=float,
    dest="fig_width",
    help="width of figure in inches"
)

ir_parser.add_argument(
    "-fh", "--figure-height",
    type=float,
    dest="fig_height",
    help="height of figure in inches"
)

ir_parser.add_argument(
    "-csv", "--experimental-csv",
    type=str,
    nargs="+",
    dest="exp_data",
    help="CSV file containing observed spectrum data, which will be plotted on top\n"
    "frequency job files should not come directly after this flag"
)

ir_parser.add_argument(
    "-rx", "--rotate-x-ticks",
    action="store_true",
    dest="rotate_x_ticks",
    default=False,
    help="rotate x-axis tick labels by 45 degrees"
)

args = ir_parser.parse_args()

if bool(args.centers) != bool(args.widths):
    sys.stderr.write(
        "both -sw/--section-widths and -sc/--section-centers must be specified"
    )
    sys.exit(2)

if args.ranges and bool(args.ranges) == bool(args.widths):
    sys.stderr.write(
        "cannot use -r/--ranges with -sw/--section-widths"
    )
    sys.exit(2)
    
centers = args.centers
widths = args.widths
if args.ranges:
    centers = []
    widths = []
    for (xmin, xmax) in args.ranges:
        centers.append((xmin + xmax) / 2)
        widths.append(abs(xmax - xmin))

exp_data = None
if args.exp_data:
    exp_data = []
    for f in args.exp_data:
        data = np.loadtxt(f, delimiter=",")
        
        for i in range(1, data.shape[1]):
            exp_data.append((data[:,0], data[:,i], None))

for f in glob_files(args.infiles, parser=ir_parser):
    fr = FileReader(f, just_geom=False)

    freq = fr.other["frequency"]

    fig = plt.gcf()
    fig.clear()

    freq.plot_ir(
        fig,
        centers=centers,
        widths=widths,
        plot_type=args.plot_type,
        peak_type=args.peak_type,
        reverse_x=args.reverse_x,
        fwhm=args.fwhm,
        point_spacing=args.point_spacing,
        voigt_mixing=args.voigt_mixing,
        linear_scale=args.linear_scale,
        quadratic_scale=args.quadratic_scale,
        exp_data=exp_data,
        anharmonic=freq.anharm_data and args.anharmonic,
        rotate_x_ticks=args.rotate_x_ticks,
    )

    if args.fig_width:
        fig.set_figwidth(args.fig_width)

    if args.fig_height:
        fig.set_figheight(args.fig_height)

    if args.outfile:
        outfile_name = args.outfile.replace("$INFILE", get_filename(f))
        plt.savefig(outfile_name, dpi=300)
    else:
        plt.show()
